from __future__ import absolute_import
from __future__ import print_function
import numpy as nm
import sys
import os
import os.path as op
import fnmatch
import shutil
import glob
from .base import output, ordered_iteritems, Struct, basestr
import six
import pickle
import warnings
import scipy.sparse as sp

try:
    import tables as pt
except:
    pt = None

class InDir(Struct):
    """
    Store the directory name a file is in, and prepend this name to other
    files.

    Examples
    --------

    >>> indir = InDir('output/file1')
    >>> print indir('file2')
    """
    def __init__(self, filename):
        self.dir = op.split(op.join(os.getcwd(), filename))[0]

    def __call__(self, filename):
        return op.join(self.dir, filename)

def ensure_path(filename):
    """
    Check if path to `filename` exists and if not, create the necessary
    intermediate directories.
    """
    dirname = os.path.dirname(filename)
    if dirname:
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        if not os.path.isdir(dirname):
            raise IOError('cannot ensure path for "%s"!' % filename)

def locate_files(pattern, root_dir=os.curdir, **kwargs):
    """
    Locate all files matching fiven filename pattern in and below
    supplied root directory.

    The `**kwargs` arguments are passed to ``os.walk()``.
    """
    for dirpath, dirnames, filenames in os.walk(os.path.abspath(root_dir),
                                                **kwargs):
        for filename in fnmatch.filter(filenames, pattern):
            yield os.path.join(dirpath, filename)

def remove_files(root_dir, **kwargs):
    """
    Remove all files and directories in supplied root directory.

    The `**kwargs` arguments are passed to ``os.walk()``.
    """
    for dirpath, dirnames, filenames in os.walk(os.path.abspath(root_dir),
                                                **kwargs):
        for filename in filenames:
            os.remove(os.path.join(root_dir, filename))

        for dirname in dirnames:
            shutil.rmtree(os.path.join(root_dir, dirname))

def remove_files_patterns(root_dir, patterns, ignores=None,
                          verbose=False):
    """
    Remove files with names satisfying the given glob patterns in a supplied
    root directory. Files with patterns in `ignores` are omitted.
    """
    from itertools import chain

    if ignores is None: ignores = []
    for _f in chain(*[glob.glob(os.path.join(root_dir, pattern))
                      for pattern in patterns]):
        can_remove = True
        for ignore in ignores:
            if fnmatch.fnmatch(_f, os.path.join(root_dir, ignore)):
                can_remove = False
                break

        if can_remove:
            output('removing "%s"' % _f, verbose=verbose)
            os.remove(_f)

def save_options(filename, options_groups, save_command_line=True,
                 quote_command_line=False):
    """
    Save groups of options/parameters into a file.

    Each option group has to be a sequence with two items: the group name and
    the options in ``{key : value}`` form.
    """
    with open(filename, 'w') as fd:
        if save_command_line:
            fd.write('command line\n')
            fd.write('------------\n\n')
            if quote_command_line:
                fd.write(' '.join('"%s"' % ii for ii in sys.argv) + '\n')

            else:
                fd.write(' '.join(sys.argv) + '\n')

        for options_group in options_groups:
            name, options = options_group
            fd.write('\n%s\n' % name)
            fd.write(('-' * len(name)) + '\n\n')
            for key, val in ordered_iteritems(options):
                fd.write('%s: %s\n' % (key, val))

def enc(string, encoding='utf-8'):
    """
    Encode given string or bytes using the specified encoding.
    """
    if sys.version_info > (3, 0):
        string = string.encode(encoding)
    return string

def dec(val, encoding='utf-8'):
    """
    Decode given bytes using the specified encoding.
    """
    if isinstance(val, bytes) and sys.version_info > (3, 0):
        val = val.decode(encoding)
    return val

##
# 27.04.2006, c
def get_trunk(filename):
    return op.splitext(op.basename(filename))[0]

def edit_filename(filename, prefix='', suffix='', new_ext=None):
    """
    Edit a file name by add a prefix, inserting a suffix in front of a file
    name extension or replacing the extension.

    Parameters
    ----------
    filename : str
        The file name.
    prefix : str
        The prefix to be added.
    suffix : str
        The suffix to be inserted.
    new_ext : str, optional
        If not None, it replaces the original file name extension.

    Returns
    -------
    new_filename : str
        The new file name.
    """
    path, filename = os.path.split(filename)
    base, ext = os.path.splitext(filename)

    if new_ext is None:
        new_filename = prefix + base + suffix + ext

    else:
        new_filename = prefix + base + suffix + new_ext

    return os.path.join(path, new_filename)

def get_print_info(n_step, fill=None):
    """
    Returns the max. number of digits in range(n_step) and the corresponding
    format string.

    Examples:

    >>> get_print_info(11)
    (2, '%2d')
    >>> get_print_info(8)
    (1, '%1d')
    >>> get_print_info(100)
    (2, '%2d')
    >>> get_print_info(101)
    (3, '%3d')
    >>> get_print_info(101, fill='0')
    (3, '%03d')
    """
    if n_step > 1:
        n_digit = int(nm.log10(n_step - 1) + 1)
        if fill is None:
            format = '%%%dd' % n_digit
        else:
            format = '%%%s%dd' % (fill, n_digit)
    else:
        n_digit, format = 0, None
    return n_digit, format

def skip_read_line(fd, no_eof=False):
    """
    Read the first non-empty line (if any) from the given file
    object. Return an empty string at EOF, if `no_eof` is False. If it
    is True, raise the EOFError instead.
    """
    ls = ''
    while 1:
        try:
            line = fd.readline()

        except EOFError:
            break

        if not line:
            if no_eof:
                raise EOFError

            else:
                break

        ls = line.strip()
        if ls and (ls[0] != '#'):
            break

    return ls

def look_ahead_line(fd):
    """
    Read and return a line from the given file object. Saves the current
    position in the file before the reading occurs and then, after the reading,
    restores the saved (original) position.
    """
    lastpos = fd.tell()
    line = fd.readline()
    fd.seek(lastpos)

    return line

def read_token(fd):
    """
    Read a single token (sequence of non-whitespace characters) from the
    given file object.

    Notes
    -----
    Consumes the first whitespace character after the token.
    """
    out = ''
    # Skip initial whitespace.

    while 1:
        ch = fd.read(1)
        if ch.isspace(): continue
        elif len(ch) == 0: return out
        else: break

    while not ch.isspace():
        out = out + ch
        ch = fd.read(1)
        if len(ch) == 0: break

    return out

def read_array(fd, n_row, n_col, dtype):
    """
    Read a NumPy array of shape `(n_row, n_col)` from the given file
    object and cast it to type `dtype`.
    If `n_col` is None, determine the number of columns automatically.
    """
    if n_col is None:
        idx = fd.tell()
        row = fd.readline().split()
        fd.seek(idx)
        n_col = len(row)

    count = n_row * n_col
    val = nm.fromfile(fd, sep=' ', count=count)

    if val.shape[0] < count:
        raise ValueError('(%d, %d) array reading failed!' % (n_row, n_col))

    val = nm.asarray(val, dtype=dtype)
    val.shape = (n_row, n_col)

    return val

##
# c: 05.02.2008, r: 05.02.2008
def read_list(fd, n_item, dtype):
    vals = []
    ii = 0
    while ii < n_item:
        line = [dtype(ic) for ic in fd.readline().split()]
        vals.append(line)
        ii += len(line)
    if ii > n_item:
        output('corrupted row?', line, ii, n_item)
        raise ValueError

    return vals

def write_dict_hdf5(filename, adict, level=0, group=None, fd=None):

    if level == 0:
        fd = pt.open_file(filename, mode='w', title='Recursive dict dump')
        group = '/'

    for key, val in six.iteritems(adict):
        if isinstance(val, dict):
            group2 = fd.create_group(group, '_' + str(key), '%s group' % key)
            write_dict_hdf5(filename, val, level + 1, group2, fd)
        else:
            if not isinstance(val, basestr):
                fd.create_array(group, '_' + str(key), val, '%s data' % key)

            else:
                fd.create_array(group, '_' + str(key), enc(val),
                                '%s data' % key)

    if level == 0:
        fd.close()

def read_dict_hdf5(filename, level=0, group=None, fd=None):
    out = {}

    if level == 0:
        fd = pt.open_file(filename, mode='r')
        group = fd.root

    for name, gr in six.iteritems(group._v_groups):
        name = name.replace('_', '', 1)
        out[name] = read_dict_hdf5(filename, level + 1, gr, fd)

    for name, data in six.iteritems(group._v_leaves):
        name = name.replace('_', '', 1)
        val = data.read()
        if isinstance(val, bytes):
            val = dec(val)

        out[name] = val

    if level == 0:
        fd.close()

    return out

##
# 02.07.2007, c
def write_sparse_matrix_hdf5(filename, mtx, name='a sparse matrix'):
    """Assume CSR/CSC."""
    with pt.open_file(filename, mode='w', title=name) as fd:
        write_sparse_matrix_to_hdf5(fd, fd.root, mtx)

def write_sparse_matrix_to_hdf5(fd, group, mtx):
    """
    Write sparse matrix to given data group of hdf5 file

    Parameters
    ----------
    group: tables.group.group
        The hdf5 file group the matrix will be read from.
    mtx: scipy.sparse.base.spmatrix
        The writed matrix
    """
    try:
        info = fd.create_group(group, 'info')
        fd.create_array(info, 'dtype', enc(mtx.dtype.str))
        fd.create_array(info, 'shape', mtx.shape)
        fd.create_array(info, 'format', enc(mtx.format))

        data = fd.create_group(group, 'data')
        fd.create_array(data, 'data', mtx.data)
        fd.create_array(data, 'indptr', mtx.indptr)
        fd.create_array(data, 'indices', mtx.indices)

    except:
        print('matrix must be in SciPy sparse CSR/CSC format!')
        print(mtx.__repr__())
        raise

##
# 02.07.2007, c
# 08.10.2007
def read_sparse_matrix_hdf5(filename, output_format=None):

    with pt.open_file(filename, mode='r') as fd:
        out = read_sparse_matrix_from_hdf5(fd, fd.root, output_format)
    return out

def read_sparse_matrix_from_hdf5(fd, group, output_format=None):
    """
    Read sparse matrix from given data group of hdf5 file

    Parameters
    ----------
    fd: tables.File
        The hdf5 file handle the matrix will be read from.
    group: tables.group.group
        The hdf5 file group of the file the matrix will be read from.
    output_format: {'csr', 'csc', None}, optional
        The resulting matrix will be in CSR or CSC format
        if this parameter is not None (which is default),
        otherwise it will be in the format the matrix was
        stored.

    Returns
    -------
    scipy.sparse.base.spmatrix
        Readed matrix
    """
    info = group.info
    data = group.data

    constructors = {'csr' : sp.csr_matrix, 'csc' : sp.csc_matrix}

    format = dec(info.format.read())
    dtype = dec(info.dtype.read())

    if output_format is None:
        constructor = constructors[format]
    else:
        constructor = constructors[output_format]

    if format in ['csc', 'csr']:
        mtx = constructor((data.data.read(),
                           data.indices.read(), data.indptr.read()),
                          shape=info.shape.read(), dtype=dtype)
    elif format == 'coo':
        mtx = constructor((data.data.read(),
                           nm.c_[data.rows.read(), data.cols.read()].T),
                          shape=info.shape.read(), dtype=dtype)
    else:
        print(format)
        raise ValueError

    if output_format in ['csc', 'csr']:
        mtx.sort_indices()

    return mtx

def path_of_hdf5_group(group):
    return group._v_pathname

class HDF5BaseData(object):
    """
    When storing values to HDF5, special classes can be used that wrap the
    stored data and modify the way the storing is done. This class is the base
    of those.
    """

    def unpack_data(self):
        """
        One can request unpacking of the wrappers during saving.

        Returns
        -------
        object
            The original object, if possible, or self.
        """
        return self

class DataMarker(HDF5BaseData):
    """
    The Base class for classes for marking data to be handled in a special way
    during saving to a HDF5 file by write_to_hdf5(). The usage is simple: just
    "decorate" the desired data element, e.g.::

        data = [data1, Cached(data2)]
        write_to_hdf5(... , ... , data)
    """

    def __init__(self, data):
        self.data = data

    def unpack_data(self):
        return self.data

class Cached(DataMarker):
    """
    The wrapper class that marks data, that should be checked
    during saving, whether it has been stored to the hdf5 file already
    and if so, a softlink to the already created instance is created
    instead of saving.
    """

class Uncached(DataMarker):
    """
    The wrapper class that marks data, that should be always stored
    to the hdf5 file, even if the object has been already stored at a
    different path in the file and so it would have been stored
    by a softlink otherwise (IGDomain, Mesh and sparse matrices behave
    so).
    """

class HDF5Data(HDF5BaseData):
    """
    Some data written to the HDF5 file can have a custom format.
    Descendants of this class should have the method `.write_data()`
    or redefine the `.write()` method.
    """

    def write(self, fd, group, name, cache=None):
        """
        Write a data structure to the HDF5 file.

        Create the following structure in the HDF5 file:
        {type: self.get_type(), anything writed by self.write_data()}

        Parameters
        ----------
        fd: tables.File
            The hdf5 file handle the data should be writed in.
        group: tables.group.Group
            The group the data will be stored to
        name: str
            Name of node that will be appended to group and will contain
            the data
        cache: dict or None, optional
            Store for already cached objects with structs id(obj) : /path/to
            Can be used for not storing the one object twice.
        """
        dgroup = fd.create_group(group, name)
        fd.create_array(dgroup, 'type', nm.array(self.get_type()))
        self.write_data(fd, dgroup, cache)
        return dgroup

    def write_data(self, fd, group):
        """
        Write data to the HDF5 file. Redefine this function in sub-classes.

        Parameters
        ----------
        fd: tables.File
            The hdf5 file handle the data should be writed to.
        group: tables.group.Group
            The group the data should be stored to.
        """
        raise Exception('Unimplemented')

class SoftLink(HDF5Data):
    """
    This object is written to the HDF5 file as a softlink to the given path.
    """
    def __init__(self, destination):
        """
        Parameters
        ----------
        destination: str
            The link destination.
        """
        self.destination = destination

    def write(self, fd, group, name, cache=None):
        """Create the softlink to the destination."""
        return fd.create_soft_link(group, name, self.destination)

class DataSoftLink(HDF5Data):
    """
    This object is written to the HDF5 file as a softlink to the given path.
    The destination of the softlink should contain only data, so
    the structure {type: type, data: softlink_to(destination)}
    is created in the place where the softlink is written.
    """

    def __init__(self, type, destination, cache=None):
        """
        Parameters
        ----------
        type: str
            Type of the object that should be written.
            See `read_from_hdf5()` for the possibilities.
        destination: str
            The Destination of the soft link where the data are stored.
        cache: object, True or None, default None
            - If None, do nothing special.
            - If True or an object cache with destination as key, storing
              such a softlink to several places does not lead to several copies
              of the object.
            - If an object is given, just identify the softlink with the
              object.
        """
        self.type = type
        self.destination = destination
        self.cache = cache

    def unpack_data(self):
        if self.cache is not None and self.cache is not True:
            return self.cache
        return self

    def get_type(self):
        return self.type

    def write_data(self, fd, group, cache=None):
        """Create the softlink to the destination and handle the caching."""

        if self.cache:
            keys = self.destination
            if self.cache is not True:
                if cache is not None:
                    cache_id = id(self.cache)
                    if cache_id in cache:
                        keys = (keys, cache[cache_id])
                    else:
                        cache[cache_id] = path_of_hdf5_group(group)

            fd.create_array(group, 'cache', nm.array(keys))
        return fd.create_soft_link(group, "data", self.destination)

def write_to_hdf5(fd, group, name, data, cache=None,
                  unpack_markers=False):
    """
    Save custom data to a HDF5 file group to be restored by read_from_hdf5().

    Allows saving lists, dicts, numpy arrays, scalars, sparse matrices,
    meshes and iga domains and all pickleable objects.

    Parameters
    ----------
    fd: tables.File
        The hdf5 file handle the data should be written in.
    group: tables.group.Group
        The group the data will be stored to.
    name: str
        The name of the node that will be appended to the group and will
        contain the data.
    data: object
        Data to be stored in the HDF5 file.
    cache: dict or None
        The cache where the paths to stored objects (currently meshes and iga
        domains) are stored, so subsequent attempts to store such objects create
        only softlinks to the initially stored object.
        The id() of objects serve as the keys into the cache.
        Mark the object with Cached() or Uncached() for (no) softlinking.
    unpack_markers:
        If True, the input data is modified so that Cached and Uncached markers
        are removed from all sub-elements of the data.

    Returns
    -------
    tables.group.Group
        The HDF5 group the data was stored to.
    """

    #imports must be done here due to circular references
    from sfepy.discrete.iga.domain import IGDomain
    from sfepy.discrete.fem.mesh import Mesh
    from sfepy.discrete.fem.meshio import HDF5MeshIO

    if cache is None: cache = {}

    def _write_to_hdf5(group, name, data):

        def save_value(type, data):
            dgroup = fd.create_group(group, name)
            fd.create_array(dgroup, 'type', nm.array(type))
            fd.create_array(dgroup, 'data', data)
            return dgroup

        def save_dict(type, data):
            dgroup = fd.create_group(group, name)
            fd.create_array(dgroup, 'type', nm.array(type))
            data_group = fd.create_group(dgroup, 'data')
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                for d in data:
                    _write_to_hdf5(data_group, d, data[d])
            if unpack_markers:
                for d in data:
                    if isinstance(data[d], HDF5BaseData):
                        data[d] = data[d].unpack_data()
            return dgroup

        def save_list(type, data):
            dgroup = fd.create_group(group, name)
            fd.create_array(dgroup, 'type', nm.array(type))
            fd.create_array(dgroup, 'len', len(data))
            data_group = fd.create_group(dgroup, 'data')

            #suppress warning that nodes with numeric ids isn't accessible as
            #python attribute
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                for i, d in enumerate(data):
                    _write_to_hdf5(data_group, str(i), d)
            if unpack_markers:
                for i,d in enumerate(data):
                    if isinstance(d, DataMarker):
                        data[i] = d.data
            return dgroup

        def save_by_function(type, fn=None, data_arg=None, cached=False):
            """
            If type is None (cached must be true), try to cache data_arg,
            otherwise save just save the data_arg. Otherwise save the data
            using the given function or just save the raw data to an array.
            """
            if cached:
                obj = data if type is not None else data_arg
                if id(obj) in cache:
                    return fd.create_soft_link(group, name,cache[id(obj)])
                if type is None:
                    #store the argument into the group
                    out = _write_to_hdf5(group, name, data_arg)
                    cache[id(data_arg)] = path_of_hdf5_group(out)
                    return out

            dgroup = fd.create_group(group, name)
            cache[id(data)] = path_of_hdf5_group(dgroup)
            fd.create_array(dgroup, 'type', nm.array(type))
            if fn is None:
                fd.create_array(dgroup, 'data', data)
            else:
                data_group = fd.create_group(dgroup, 'data')
                if data_arg is None:
                    fn(fd, data_group)
                else:
                    fn(fd, data_group, data_arg)
            return dgroup

        def save_type(type):
            dgroup = fd.create_group(group, name)
            fd.create_array(dgroup, 'type', nm.array(type))
            return dgroup

        for t in (True, False, None):
            if data is t:
                return save_type(str(t))

        if isinstance(data, dict):
            return save_dict('dict', data)

        if isinstance(data, list):
            return save_list('list', data)

        if isinstance(data, tuple):
            return save_list('tuple', data)

        if isinstance(data, (int, float, complex, nm.ndarray)):
            return save_value('raw', data)

        if isinstance(data, str):
            return save_value('str', nm.array(enc(data)))

        if isinstance(data, (sp.csr_matrix, sp.csc_matrix)):
            return save_by_function('sparse_matrix',
                                    write_sparse_matrix_to_hdf5,
                                    data, cached=True)

        if isinstance(data, IGDomain):
            return save_by_function('IGDomain', data.write_domain_to_hdf5,
                                    cached=True)

        if isinstance(data, Mesh):
            return save_by_function('Mesh', HDF5MeshIO.write_mesh_to_hdf5,
                                    data_arg=data, cached=True)

        if isinstance(data, Uncached):
            return write_to_hdf5(fd, group, name, data.data)

        if isinstance(data, Cached):
            return save_by_function(None, None, data.data, cached=True)

        if isinstance(data, HDF5Data):
            return data.write(fd, group, name, cache)

        if isinstance(data, Struct):
            return save_dict('Struct', data.to_dict())

        return save_value('pickle', nm.array(pickle.dumps(data)))

    return _write_to_hdf5(group, name, data)

def read_from_hdf5(fd, group, cache=None):
    """
    Read custom data from a HDF5 file group saved by write_to_hdf5().

    The data are stored in a general (possibly nested) structure:
    {
      'type' : string type identificator
      'data' : stored data
      'cache': string, optional - another posible location of object
    }

    Parameters
    ----------
    fd: tables.File
        The hdf5 file handle the data should be restored from.
    group: tables.group.Group
        The group in the hdf5 file the data will be restored from.
    cache: dict or None
        Some objects (e.g. Mesh instances) can be stored on more places in the
        HDF5 file tree using softlinks, so when the data are restored, the
        restored objects are stored and searched in cache so that they are
        created only once. The keys to cache are the (real) paths of the
        created objects.
        Moreover, if some stored object has a 'cache' key (see e.g.
        DataSoftLink class), and the object with a given 'path' has been
        already created, it is returned instead of creating a new object.
        Otherwise, the newly created object is associated both with its real
        path and with the cache key path.

        The caching is not active for scalar data types.

    Returns
    -------
    data : object
        The restored custom data.
    """
    if cache is None:
       cache = {}

    from sfepy.discrete.iga.domain import IGDomain
    from sfepy.discrete.fem.meshio import HDF5MeshIO
    types = {b'True' : True, b'False' : False, b'None' : None}

    def _read_from_hdf5(group):
        while isinstance(group, pt.link.SoftLink):
            group = group()
        path = path_of_hdf5_group(group)
        if path in cache:
            return cache[path]

        cache_ids = tuple()
        if 'cache' in group:
            cache_ids = group.cache.read().reshape(-1)
            cache_ids = [dec(i) for i in cache_ids]
            #for DataSoftlink, there can be 2 paths, one for the softlink target
            #and one for the object identified with the softlink
            for cache_id in cache_ids:
                if cache_id in cache:
                    out = cache[cache_id]
                    break
            else:
                out = _read_value_from_hdf5(group)
        else:
            out = _read_value_from_hdf5(group)
        if not isinstance(out, (int, float, str, bool, None.__class__)):
           cache[path] = out
           for cache_id in cache_ids:
               cache[cache_id] = out
        return out

    def _read_value_from_hdf5(group):

        type = group.type.read().item()
        if type in types:
            return types[type]

        data = group.data
        while isinstance(data, pt.link.SoftLink):
            data = data()

        def load_list():
            out = [None] * group.len.read()
            for i in data:
                out[int(i._v_name)] = _read_from_hdf5(i)
            return out

        def load_dict():
            out = {}
            for i in fd.iter_nodes(data):
                out[i._v_name] = _read_from_hdf5(i)
            return out

        def load_object(fn=None):
            if fn:
                out = fn(fd, data)
            else:
                out = data.read()
            return out

        if type == b'raw':
            return data.read()

        if type == b'object':
            return load_object()

        if type == b'str':
            return dec(data.read().item())

        if type == b'pickle':
            return pickle.loads(data.read().item())

        if type == b'dict':
            return load_dict()

        if type == b'Struct':
            return Struct(**load_dict())

        if type == b'list':
            return load_list()

        if type == b'tuple':
            return tuple(load_list())

        if type == b'sparse_matrix':
            return load_object(read_sparse_matrix_from_hdf5)

        if type == b'IGDomain':
            return load_object(IGDomain.read_domain_from_hdf5)

        if type == b'Mesh':
            return load_object(HDF5MeshIO.read_mesh_from_hdf5)

        raise Exception('Unknown h5 group type {}'.format(type.decode('utf8')))

    return _read_from_hdf5(group)

class HDF5ContextManager:
    def __init__(self, filename, *args, **kwargs):
        self.filename = filename
        self.file = None
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        if isinstance(self.filename, pt.File):
           return self.filename
        else:
           self.file = pt.open_file(self.filename, *self.args, **self.kwargs)
           return self.file

    def __exit__(self, type, value, traceback):
        if self.file:
           self.file.close()
           self.file = None

def get_or_create_hdf5_group(fd, path, from_group=None):
    if from_group is None:
       from_group = fd.root
    if path == '':
       return from_group
    if path[0] == '/':
       path = path[1:]
    for name in path.split('/'):
       if name in from_group:
           from_group = getattr(from_group, name)
       else:
           from_group = pt.create_group(from_group, name)
    return from_group



r"""
Elastic contact sphere simulating an indentation test.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    + \int_{\Gamma} \ul{v} \cdot f(d(\ul{u})) \ul{n}(\ul{u})
    = 0 \;,

where

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl} + \delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl}
    \;.

Notes
-----

Even though the material is linear elastic and small deformations are used, the
problem is highly nonlinear due to contacts with the sphere. See also
elastic_contact_planes.py example.
"""
from __future__ import absolute_import
from sfepy import data_dir
from sfepy.mechanics.matcoefs import stiffness_from_lame

filename_mesh = data_dir + '/meshes/3d/cube_medium_hexa.mesh'

k = 1e5 # Elastic sphere stiffness for positive penetration.
f0 = 1e-2 # Force at zero penetration.

options = {
    'nls' : 'newton',
    'ls' : 'ls',

    'output_format': 'vtk',
}

fields = {
    'displacement': ('real', 3, 'Omega', 1),
}

materials = {
    'solid' : ({
        'D': stiffness_from_lame(dim=3, lam=5.769, mu=3.846),
    },),
    'cs' : ({
        'f' : [k, f0],
        '.c' : [0.0, 0.0, 1.2],
        '.r' : 0.8,
    },),
}

variables = {
    'u' : ('unknown field', 'displacement', 0),
    'v' : ('test field', 'displacement', 'u'),
}

regions = {
    'Omega' : 'all',
    'Bottom' : ('vertices in (z < -0.499)', 'facet'),
    'Top' : ('vertices in (z > 0.499)', 'facet'),
}

ebcs = {
    'fixed' : ('Bottom', {'u.all' : 0.0}),
}

equations = {
    'elasticity' :
    """dw_lin_elastic.2.Omega(solid.D, v, u)
     + dw_contact_sphere.2.Top(cs.f, cs.c, cs.r, v, u)
     = 0""",
}

solvers = {
    'ls' : ('ls.scipy_direct', {}),
    'newton' : ('nls.newton', {
        'i_max' : 20,
        'eps_a' : 1e-1,
        'ls_on' : 2.0,
        'check' : 0,
        'delta' : 1e-6,
    }),
}

def main():
    import os

    import numpy as nm
    import matplotlib.pyplot as plt

    from sfepy.discrete.fem import MeshIO
    import sfepy.linalg as la
    from sfepy.mechanics.contact_bodies import ContactSphere, plot_points

    conf_dir = os.path.dirname(__file__)
    io = MeshIO.any_from_filename(filename_mesh, prefix_dir=conf_dir)
    bb = io.read_bounding_box()
    outline = [vv for vv in la.combine(zip(*bb))]

    ax = plot_points(None, nm.array(outline), 'r*')
    csc = materials['cs'][0]
    cs = ContactSphere(csc['.c'], csc['.r'])

    pps = (bb[1] - bb[0]) * nm.random.rand(5000, 3) + bb[0]
    mask = cs.mask_points(pps, 0.0)

    ax = plot_points(ax, cs.centre[None, :], 'b*', ms=30)
    ax = plot_points(ax, pps[mask], 'kv')
    ax = plot_points(ax, pps[~mask], 'r.')

    plt.show()

if __name__ == '__main__':
    main()



from __future__ import absolute_import
import numpy as nm

from sfepy.base.testing import TestCommon

class Test(TestCommon):

    @staticmethod
    def from_conf(conf, options):
        test = Test(conf=conf, options=options)
        return test

    def test_stokes_slip_bc(self):
        import scipy.sparse as sp

        from sfepy.base.conf import ProblemConf
        from sfepy.discrete import Problem
        import examples.navier_stokes.stokes_slip_bc as ssb

        conf = ProblemConf.from_module(ssb)
        pb = Problem.from_conf(conf, init_solvers=False)
        pb.time_update()
        variables = pb.get_variables()

        adi = variables.adi
        lcdi = variables.lcdi
        mtx = variables.mtx_lcbc

        ok = adi.var_names == lcdi.var_names
        self.report('same adi-lcdi ordering:', ok)

        ublock = mtx[adi.indx['u']]
        ir, ic = ublock.nonzero()
        ir += adi.indx['u'].start

        i0, i1 = adi.indx['u'].start, adi.indx['u'].stop
        _ok0 = (i0 <= ir).all() and  (ir < i1).all()
        self.report('u block rows in [%d %d[: %s' % (i0, i1, _ok0))

        i0, i1 = lcdi.indx['u'].start, lcdi.indx['u'].stop
        _ok1 = (i0 <= ic).all() and  (ic < i1).all()
        self.report('u block cols in [%d %d[: %s' % (i0, i1, _ok1))

        ok = ok and _ok0 and _ok1

        pblock = mtx[adi.indx['p']]
        ir, ic, iv = sp.find(pblock)
        ir += adi.indx['p'].start

        i0, i1 = adi.indx['p'].start, adi.indx['p'].stop
        _ok0 = (i0 <= ir).all() and  (ir < i1).all()
        self.report('p block rows in [%d %d[: %s' % (i0, i1, _ok0))

        i0, i1 = lcdi.indx['p'].start, lcdi.indx['p'].stop
        _ok1 = (i0 <= ic).all() and  (ic < i1).all()
        self.report('p block cols in [%d %d[: %s' % (i0, i1, _ok1))

        ok = ok and _ok0 and _ok1

        _ok0 = (len(ir) == adi.n_dof['p'])
        self.report('p block size correct:', _ok0)
        _ok1 = ((ir - adi.indx['p'].start) == (ic - lcdi.indx['p'].start)).all()
        self.report('p block diagonal:', _ok1)
        _ok2 = (iv == 1.0).all()
        self.report('p block identity:', _ok2)

        ok = ok and _ok0 and _ok1 and _ok2

        return ok

    def test_laplace_shifted_periodic(self):
        from sfepy.mesh.mesh_generators import gen_block_mesh
        from sfepy.discrete.fem import FEDomain
        from examples.diffusion.laplace_shifted_periodic import run

        dims = [2.0, 1.0]
        shape = [21, 11]
        centre = [0.0, 0.0]
        mesh = gen_block_mesh(dims, shape, centre, name='block-fem')
        fe_domain = FEDomain('domain', mesh)

        pb, state = run(fe_domain, 3)

        gamma3 = pb.domain.regions['Gamma3']
        gamma4 = pb.domain.regions['Gamma4']

        field = pb.fields['fu']

        # Check that the shift equals to one.
        i3 = field.get_dofs_in_region(gamma3, merge=True)
        i4 = field.get_dofs_in_region(gamma4, merge=True)

        i_corners = nm.array([0, shape[0] - 1])
        ii = nm.setdiff1d(nm.arange(len(i3)), i_corners)

        vals = state()

        shift = vals[i3] - vals[i4]

        ok = (shift[i_corners] == 0.0).all()

        ok = ok and nm.allclose(shift[ii], 1.0, rtol=0.0, atol=1e-14)

        if not ok:
            self.report('wrong shift:', shift)

        return ok



from __future__ import absolute_import
import sfepy
from . import terms
from . import extmods
from .terms import Terms, Term
from .terms_th import THTerm, ETHTerm
from sfepy.base.base import load_classes

term_files = sfepy.get_paths('sfepy/terms/terms*.py')
term_table = load_classes(term_files, [Term], ignore_errors=True)

del sfepy

def register_term(cls):
    """
    Register a custom term.
    """
    term_table[cls.name] = cls



"""
Functions implementing the shell10x element.
"""
from __future__ import absolute_import
import numpy as nm

from sfepy.linalg import norm_l2_along_axis as norm
from sfepy.linalg import dot_sequences as ddot
from six.moves import range

def create_elastic_tensor(young, poisson, shear_correction=True):
    """
    Create the elastic tensor with the applied shear correction (the default)
    for the shell10x element.
    """
    from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson

    # Static condensation so that \sigma_{33} = 0.
    mtx = stiffness_from_youngpoisson(3, young, poisson, plane='stress')
    mtx[2, :3] = mtx[:2, 2] = 0.0

    if shear_correction:
        mtx[4, 4] *= 5.0 / 6.0
        mtx[5, 5] *= 5.0 / 6.0

    return mtx

def create_transformation_matrix(coors):
    r"""
    Create a transposed coordinate transformation matrix, that transforms 3D
    coordinates of quadrilateral cell vertices so that the transformed vertices
    of a plane cell are in the :math:`x-y` plane. The rotation is performed
    w.r.t. the centres of quadrilaterals.

    Parameters
    ----------
    coors : array
        The coordinates of cell vertices, shape `(n_el, 4, 3)`.

    Returns
    -------
    mtx_t : array
        The transposed transformation matrix :math:`T`, i.e.
        :math:`X_{inplane} = T^T X_{3D}`.

    Notes
    -----
    :math:`T = [t_1, t_2, n]`, where :math:`t_1`, :math:`t_2`, are unit
    in-plane (column) vectors and :math:`n` is the unit normal vector,
    all mutually orthonormal.
    """
    # Local coordinate system.
    t1 = coors[:, 1, :] + coors[:, 2, :] - coors[:, 0, :] - coors[:, 3, :]
    t2 = coors[:, 2, :] + coors[:, 3, :] - coors[:, 0, :] - coors[:, 1, :]
    n = nm.cross(t1, t2)
    t2 = nm.cross(n, t1)

    t1 = t1 / norm(t1)[:, None]
    t2 = t2 / norm(t2)[:, None]
    n = n / norm(n)[:, None]

    # Coordinate transformation matrix (transposed!).
    mtx_t = nm.concatenate((t1[:, :, None],
                            t2[:, :, None],
                            n[:, :, None]), axis=2)

    return mtx_t

def transform_asm_matrices(out, mtx_t, blocks):
    """
    Transform matrix assembling contributions to global coordinate system, one
    node at a time.

    Parameters
    ----------
    out : array
        The array of matrices, transformed in-place.
    mtx_t : array
        The array of transposed transformation matrices :math:`T`, see
        :func:`create_transformation_matrix`.
    blocks : array
        The DOF blocks that are
    """
    for ir in blocks:
        ir = ir[:, None]
        for ic in blocks:
            fn = out[:, 0, ir, ic]
            out[:, 0, ir, ic] = ddot(ddot(mtx_t, fn, 'AB'), mtx_t, 'ABT')

def create_local_bases(coors):
    """
    Create local orthonormal bases in each vertex of quadrilateral cells.

    Parameters
    ----------
    coors : array
        The coordinates of cell vertices, shape `(n_el, 4, 3)`.

    Returns
    -------
    ebs : array
        The local bases, shape `(n_el, 4, 3, 3)`. The basis vectors are rows of
        the (..., 3, 3) blocks.
    """
    ebs = nm.zeros((coors.shape[0], 4, 3, 3), dtype=nm.float64)
    ebs[:, 0, 0, :] = coors[:, 1, :] - coors[:, 0, :]
    ebs[:, 0, 1, :] = coors[:, 3, :] - coors[:, 0, :]
    ebs[:, 0, 2, :] = nm.cross(ebs[:, 0, 0, :], ebs[:, 0, 1, :])
    ebs[:, 1, 0, :] = coors[:, 1, :] - coors[:, 0, :]
    ebs[:, 1, 1, :] = coors[:, 2, :] - coors[:, 1, :]
    ebs[:, 1, 2, :] = nm.cross(ebs[:, 1, 0, :], ebs[:, 1, 1, :])
    ebs[:, 2, 0, :] = coors[:, 2, :] - coors[:, 3, :]
    ebs[:, 2, 1, :] = coors[:, 2, :] - coors[:, 1, :]
    ebs[:, 2, 2, :] = nm.cross(ebs[:, 2, 0, :], ebs[:, 2, 1, :])
    ebs[:, 3, 0, :] = coors[:, 2, :] - coors[:, 3, :]
    ebs[:, 3, 1, :] = coors[:, 3, :] - coors[:, 0, :]
    ebs[:, 3, 2, :] = nm.cross(ebs[:, 3, 0, :], ebs[:, 3, 1, :])

    e2 = nm.array([0, 1, 0], dtype=nm.float64)
    for ii in range(4):
        ebs[:, ii, 0, :] = nm.cross(e2, ebs[:, ii, 2, :])
        ebs[:, ii, 1, :] = nm.cross(ebs[:, ii, 2, :], ebs[:, ii, 0, :])
        ebs[:, ii] = ebs[:, ii] / nm.linalg.norm(ebs[:, ii], axis=2)[..., None]

    return ebs

def create_rotation_ops(ebs):
    """
    Create operators associated to rotation DOFs.

    Parameters
    ----------
    ebs : array
        The local bases, shape `(n_el, 4, 3, 3)`.

    Returns
    -------
    rops : array
        The rotation operators, shape `(n_el, 4, 3, 3)`.
    """
    rops = nm.empty((ebs.shape[0], 4, 3, 3), dtype=nm.float64)
    for ii in range(4):
        aux = nm.einsum('cj,ck->cjk', ebs[:, ii, 0], ebs[:, ii, 1])
        rops[:, ii] = aux - aux.transpose((0, 2, 1))

    return rops

def create_strain_transform(mtx_ts):
    r"""
    Create strain tensor transformation matrices, given coordinate
    transformation matrices.

    Notes
    -----
    Expresses :math:`T E T^T` in terms of symmetrix storage as :math:`Q e`,
    with the ordering of components:
    :math:`e = [e_{11}, e_{22}, e_{33}, 2 e_{12}, 2 e_{13}, 2 e_{23}]`.
    """
    mtx_qs = nm.empty((mtx_ts.shape[0], mtx_ts.shape[1], 6, 6),
                      dtype=nm.float64)
    for ie in range(mtx_ts.shape[0]):
        for iq in range(mtx_ts.shape[1]):
            l = mtx_ts[ie, iq, :, 0]
            m = mtx_ts[ie, iq, :, 1]
            n = mtx_ts[ie, iq, :, 2]

            mtx_q = [
                [l[0]**2,     m[0]**2,     n[0]**2,     l[0]*m[0],           n[0]*l[0],           m[0]*n[0]],
                [l[1]**2,     m[1]**2,     n[1]**2,     l[1]*m[1],           n[1]*l[1],           m[1]*n[1]],
                [l[2]**2,     m[2]**2,     n[2]**2,     l[2]*m[2],           n[2]*l[2],           m[2]*n[2]],
                [2*l[0]*l[1], 2*m[0]*m[1], 2*n[0]*n[1], l[0]*m[1]+l[1]*m[0], n[0]*l[1]+n[1]*l[0], m[0]*n[1]+m[1]*n[0]],
                [2*l[2]*l[0], 2*m[2]*m[0], 2*n[2]*n[0], l[2]*m[0]+l[0]*m[2], n[2]*l[0]+n[0]*l[2], m[2]*n[0]+m[0]*n[2]],
                [2*l[1]*l[2], 2*m[1]*m[2], 2*n[1]*n[2], l[1]*m[2]+l[2]*m[1], n[1]*l[2]+n[2]*l[1], m[1]*n[2]+m[2]*n[1]],
                ]
            mtx_qs[ie, iq] = mtx_q

    return mtx_qs

def get_mapping_data(ebs, rops, ps, coors_loc, qp_coors, qp_weights,
                     special_dx3=False):
    r"""
    Compute reference element mapping data for shell10x elements.

    Notes
    -----
    The code assumes that the quadrature points are w.r.t. (:math:`t` =
    thickness of the shell) :math:`[0, 1] \times [0, 1] \times [-t/2, t/2]`
    reference cell and the quadrature weights are multiplied by :math:`t`.
    """
    n_el = coors_loc.shape[0]
    n_qp = qp_weights.shape[0]

    bfu = ps.eval_base(qp_coors[:, :2].copy())
    bfgu = ps.eval_base(qp_coors[:, :2].copy(), diff=True)

    nh = ebs[..., -1, :]

    # h = tilde dzeta = t * (dzeta - 0.5), d = dzeta in [0, 1]
    # dh/dd = t
    h = qp_coors[:, -1, None, None]

    dxdxi = nm.empty((n_el, n_qp, 3, 3), dtype=nm.float64)
    coors_loc_3d = coors_loc[:, None, ...] + (nh[:, None, ...] * h)
    dxdxi[..., :2, :] = nm.einsum('qij,cqjk->cqik', bfgu, coors_loc_3d)

    # The factor of t from dh/dd is not here because quadrature weights were
    # multiplied by t. This seems more accurate than having t here and the
    # usual weights.
    # dx_col/dxi_row
    dxdxi[..., 2:, :] = nm.einsum('qij,cjk->cqik', bfu, nh) # * 2
    if special_dx3:
        dxdxi[..., 2:, :] = [0.0, 0.0, 1.0]

    det = nm.linalg.det(dxdxi)

    # *2 to have [-1, 1] reference cell numbers...
    # dxi_col/dx_row
    dxidx = nm.linalg.inv(dxdxi)

    if special_dx3:
        return dxidx, det

    bfg = nm.zeros((n_el, n_qp, 3, 3, 24), dtype=nm.float64)

    bfg[..., 0, :2, :4] = bfgu
    bfg[..., 1, :2, 4:8] = bfgu
    bfg[..., 2, :2, 8:12] = bfgu

    n = None

    # bfg[:, 0, :2, 12:16] = (bfgu * rops[..., 0, 0][:, n, n, :]) * h = 0
    bfg[..., 0, :2, 16:20] = (bfgu * rops[..., 0, 1][:, n, n, :]) * h
    bfg[..., 0, :2, 20:24] = (bfgu * rops[..., 0, 2][:, n, n, :]) * h
    # bfg[:, 0, 2:, 12:16] = bfu * rops[..., 0, 0][:, n, n, :] = 0
    bfg[..., 0, 2:, 16:20] = bfu * rops[..., 0, 1][:, n, n, :]
    bfg[..., 0, 2:, 20:24] = bfu * rops[..., 0, 2][:, n, n, :]

    bfg[..., 1, :2, 12:16] = (bfgu * rops[..., 1, 0][:, n, n, :]) * h
    # bfg[:, 1, :2, 16:20] = (bfgu * rops[..., 1, 1][:, n, n, :]) * h = 0
    bfg[..., 1, :2, 20:24] = (bfgu * rops[..., 1, 2][:, n, n, :]) * h
    bfg[..., 1, 2:, 12:16] = bfu * rops[..., 1, 0][:, n, n, :]
    # bfg[:, 1, 2:, 16:20] = bfu * rops[..., 1, 1][:, n, n, :] = 0
    bfg[..., 1, 2:, 20:24] = bfu * rops[..., 1, 2][:, n, n, :]

    bfg[..., 2, :2, 12:16] = (bfgu * rops[..., 2, 0][:, n, n, :]) * h
    bfg[..., 2, :2, 16:20] = (bfgu * rops[..., 2, 1][:, n, n, :]) * h
    # bfg[:, 2, :2, 20:24] = (bfgu * rops[..., 2, 2][:, n, n, :]) * h = 0
    bfg[..., 2, 2:, 12:16] = bfu * rops[..., 2, 0][:, n, n, :]
    bfg[..., 2, 2:, 16:20] = bfu * rops[..., 2, 1][:, n, n, :]
    # bfg[:, 2, 2:, 20:24] = bfu * rops[..., 2, 2][:, n, n, :] = 0

    # ... related to [0, 1] vs. [-1, 1]:
    # bfg[:, :, 2, 12:] /= 0.5 when *2 in dxidx above

    bfgm = nm.einsum('cqij,cqkjl->cqikl', dxidx, bfg)

    return coors_loc_3d, bfu, bfgm, dxidx, det

def get_dsg_strain(coors_loc, qp_coors):
    r"""
    Compute DSG strain components.

    Returns
    -------
    dsg : array
        The strain matrix components corresponding to :math:`e_{13}`,
        :math:`e_{23}`, shape `(n_el, n_qp, 2, 24)`.

    Notes
    -----
    Involves :math:`w`, :math:`\alpha`, :math:`\beta` DOFs.
    """
    dsg = nm.zeros((coors_loc.shape[0], qp_coors.shape[0], 2, 24),
                   dtype=nm.float64)

    x1, x2, x3, x4 = coors_loc.transpose((1, 0, 2))

    o = nm.ones((coors_loc.shape[0], 1), dtype=nm.float64)

    v13_1 = nm.c_[-o, o, -0.5 * (x2[:, 1] - x1[:, 1]),
                   -0.5 * (x2[:, 1] - x1[:, 1]),
                   0.5 * (x2[:, 0] - x1[:, 0]),
                   0.5 * (x2[:, 0] - x1[:, 0])]
    i13_1 = nm.r_[8, 9, 12, 13, 16, 17]
    v13_2 = nm.c_[o, -o, -0.5 * (x3[:, 1] - x4[:, 1]),
                  -0.5 * (x3[:, 1] - x4[:, 1]),
                  0.5 * (x3[:, 0] - x4[:, 0]),
                  0.5 * (x3[:, 0] - x4[:, 0])]
    i13_2 = nm.r_[10, 11, 14, 15, 18, 19]

    v23_1 = nm.c_[-o, o, -0.5 * (x4[:, 1] - x1[:, 1]),
                   -0.5 * (x4[:, 1] - x1[:, 1]),
                   0.5 * (x4[:, 0] - x1[:, 0]),
                   0.5 * (x4[:, 0] - x1[:, 0])]
    i23_1 = nm.r_[8, 11, 12, 15, 16, 19]
    v23_2 = nm.c_[-o, o, -0.5 * (x3[:, 1] - x2[:, 1]),
                   -0.5 * (x3[:, 1] - x2[:, 1]),
                   0.5 * (x3[:, 0] - x2[:, 0]),
                   0.5 * (x3[:, 0] - x2[:, 0])]
    i23_2 = nm.r_[9, 10, 13, 14, 17, 18]

    c1, c2 = qp_coors[:, 0], qp_coors[:, 1]

    # Assuming qp in [-1, 1]
    # dsg[:, 0, i13_1] = 0.25 * (1.0 - c2)[:, None] * v13_1
    # dsg[:, 0, i13_2] = 0.25 * (1.0 + c2)[:, None] * v13_2
    # dsg[:, 1, i23_1] = 0.25 * (1.0 - c1)[:, None] * v23_1
    # dsg[:, 1, i23_2] = 0.25 * (1.0 + c1)[:, None] * v23_2

    # Assuming qp in [0, 1]
    n = None
    dsg[..., 0, i13_1] = (1.0 - c2)[n, :, n] * v13_1[:, n, :]
    dsg[..., 0, i13_2] = (c2)[n, :, n] * v13_2[:, n, :]
    dsg[..., 1, i23_1] = (1.0 - c1)[n, :, n] * v23_1[:, n, :]
    dsg[..., 1, i23_2] = (c1)[n, :, n] * v23_2[:, n, :]

    return dsg

def create_strain_matrix(bfgm, dxidx, dsg):
    """
    Create the strain operator matrix.
    """
    tg = create_strain_transform(dxidx)

    mtx_b = nm.concatenate((bfgm[..., 0:1, 0, :],
                            bfgm[..., 1:2, 1, :],
                            bfgm[..., 2:3, 2, :],
                            bfgm[..., 0:1, 1, :] + bfgm[..., 1:2, 0, :],
                            nm.einsum('ciaj,cijk->ciak', tg[..., 4:5, 4:], dsg),
                            nm.einsum('ciaj,cijk->ciak', tg[..., 5:6, 4:], dsg)),
                           2)

    return mtx_b

def add_eas_dofs(mtx_b, qp_coors, det, det0, dxidx0):
    """
    Add additional strain components [Andelfinger and Ramm] (7 parameters to be
    condensed out).
    """
    tg0 = create_strain_transform(dxidx0)

    # mtx_b is the same as Bk -> mtx_ee must be the same as Ba
    # -> qp coors transformation below is needed.
    qp_coors = 8 * (qp_coors - 0.5)

    c1, c2 = qp_coors[:, 0], qp_coors[:, 1]
    cc = c1 * c2

    mtx_e = nm.zeros((qp_coors.shape[0], 6, 7), dtype=nm.float64)

    mtx_e[:, 0, 0] = mtx_e[:, 3, 2] = c1
    mtx_e[:, 1, 1] = mtx_e[:, 3, 3] = c2
    mtx_e[:, 0, 4] = mtx_e[:, 1, 5] = mtx_e[:, 3, 6] = cc

    dd = (det0 / det)[..., None, None]
    mtx_ee = dd * nm.einsum('cij,qjk->cqik', tg0[:, 0], mtx_e)

    mtx_be = nm.concatenate((mtx_b, mtx_ee), 3)

    return mtx_be

def rotate_elastic_tensor(mtx_d, bfu, ebs):
    """
    Rotate the elastic tensor into the local coordinate system of each cell.
    The local coordinate system results from interpolation of `ebs` with the
    bilinear basis.
    """
    # ! mtx_ts is not orthonormal for non-planar cells!
    mtx_ts = nm.einsum('qi,cijk->cqjk', bfu[:, 0, :], ebs)
    # print nm.einsum('cqji,cqjk->cqik', mtx_ts, mtx_ts)

    # ? double entries here?
    tt = create_strain_transform(mtx_ts)

    mtx_dr = ddot(ddot(tt, mtx_d, 'ATB'), tt, 'AB')
    return mtx_dr

def create_drl_transform(ebs):
    """
    Create the transformation matrix for locking of the drilling rotations.
    """
    n_el = ebs.shape[0]
    mtx_drl = nm.tile(nm.eye(24, dtype=nm.float64), (n_el, 1, 1))
    for ii in range(4):
        nh = ebs[:, ii, -1, :]
        mtx = nm.c_[nm.c_[1.0 - nh[:, 0]**2,
                          -nh[:, 0]*nh[:, 1],
                          -nh[:, 0]*nh[:, 2]],
                    nm.c_[-nh[:, 0]*nh[:, 1],
                           1.0 - nh[:, 1]**2,
                           -nh[:, 1]*nh[:, 2]],
                    nh].reshape(n_el, 3, 3)
        mtx_drl[:, 12+ii:12+ii+9:4, 12+ii:12+ii+9:4] = mtx

    return mtx_drl

def lock_drilling_rotations(mtx, ebs, coefs):
    """
    Lock the drilling rotations in the stiffness matrix.
    """
    mtx_drl = create_drl_transform(ebs)
    mtx_idrl = nm.linalg.inv(mtx_drl)
    mtx_tr = ddot(ddot(mtx_drl, mtx), mtx_idrl)

    idrl = nm.arange(20, 24)
    mtx_tr[:, idrl, idrl] = coefs[:, None]

    mtx2 = ddot(ddot(mtx_idrl, mtx_tr), mtx_drl)

    return mtx2



"""
Reference-physical domain mappings.
"""
import numpy as nm

from sfepy.base.base import Struct

class PhysicalQPs(Struct):
    """
    Physical quadrature points in a region.
    """

    def __init__(self, num=0):
        Struct.__init__(self, num=num, shape=(0, 0, 0))
        self.values = nm.empty(self.shape, dtype=nm.float64)

    def get_shape(self, rshape):
        """
        Get shape from raveled shape.
        """
        n_qp = self.shape[1]

        if n_qp > 0:
            if (rshape[0] / n_qp) * n_qp != rshape[0]:
                raise ValueError('incompatible shapes! (n_qp: %d, %s)'
                                 % (n_qp, rshape))

            shape = (rshape[0] // n_qp, n_qp) + rshape[1:]

        else:
            shape = (rshape[0], 0, 0, 0)

        return shape

class Mapping(Struct):
    """
    Base class for mappings.
    """

    @staticmethod
    def from_args(region, kind='v'):
        """
        Create mapping from reference to physical entities in a given
        region, given the integration kind ('v' or 's').

        This mapping can be used to compute the physical quadrature
        points.

        Parameters
        ----------
        region : Region instance
            The region defining the entities.
        kind : 'v' or 's'
            The kind of the entities: 'v' - cells, 's' - facets.

        Returns
        -------
        mapping : VolumeMapping or SurfaceMapping instance
            The requested mapping.
        """
        from sfepy.discrete.fem.domain import FEDomain
        from sfepy.discrete.iga.domain import IGDomain

        if isinstance(region.domain, FEDomain):
            import sfepy.discrete.fem.mappings as mm
            coors = region.domain.get_mesh_coors()
            if kind == 's':
                coors = coors[region.vertices]

            conn, gel = region.domain.get_conn(ret_gel=True)

            if kind == 'v':
                cells = region.get_cells()

                mapping = mm.VolumeMapping(coors, conn[cells], gel=gel)

            elif kind == 's':
                from sfepy.discrete.fem.fe_surface import FESurface

                aux = FESurface('aux', region, gel.get_surface_entities(),
                                conn)
                mapping = mm.SurfaceMapping(coors, aux.leconn,
                                            gel=gel.surface_facet)

        elif isinstance(region.domain, IGDomain):
            import sfepy.discrete.iga.mappings as mm
            mapping = mm.IGMapping(region.domain, region.cells)

        else:
            raise ValueError('unknown domain class! (%s)' % type(region.domain))

        return mapping

def get_physical_qps(region, integral, map_kind=None):
    """
    Get physical quadrature points corresponding to the given region
    and integral.
    """
    phys_qps = PhysicalQPs()

    if map_kind is None:
        map_kind = 'v' if region.can_cells else 's'

    gmap = Mapping.from_args(region, map_kind)

    gel = gmap.get_geometry()
    qp_coors, _ = integral.get_qp(gel.name)

    qps = gmap.get_physical_qps(qp_coors)
    n_el, n_qp = qps.shape[0], qps.shape[1]

    phys_qps.num = n_el * n_qp
    phys_qps.shape = qps.shape

    qps.shape = (phys_qps.num, qps.shape[2])
    phys_qps.values = qps

    return phys_qps

def get_mapping_data(name, field, integral, region=None, integration='volume'):
    """
    General helper function for accessing reference mapping data.

    Get data attribute `name` from reference mapping corresponding to
    `field` in `region` in quadrature points of the given `integral` and
    `integration` type.

    Parameters
    ----------
    name : str
        The reference mapping attribute name.
    field : Field instance
        The field defining the reference mapping.
    integral : Integral instance
        The integral defining quadrature points.
    region : Region instance, optional
        If given, use the given region instead of `field` region.
    integration : one of ('volume', 'surface', 'surface_extra')
        The integration type.

    Returns
    -------
    data : array
        The required data merged for all element groups.

    Notes
    -----
    Assumes the same element geometry in all element groups of the field!
    """
    data = None
    if region is None:
        region = field.region

    geo, _ = field.get_mapping(region, integral, integration)
    data = getattr(geo, name)

    return data

def get_jacobian(field, integral, region=None, integration='volume'):
    """
    Get the jacobian of reference mapping corresponding to `field`.

    Parameters
    ----------
    field : Field instance
        The field defining the reference mapping.
    integral : Integral instance
        The integral defining quadrature points.
    region : Region instance, optional
        If given, use the given region instead of `field` region.
    integration : one of ('volume', 'surface', 'surface_extra')
        The integration type.

    Returns
    -------
    jac : array
        The jacobian merged for all element groups.

    See Also
    --------
    get_mapping_data

    Notes
    -----
    Assumes the same element geometry in all element groups of the field!
    """
    jac = get_mapping_data('det', field, integral, region=region,
                           integration=integration)
    return jac

def get_normals(field, integral, region):
    """
    Get the normals of element faces in `region`.

    Parameters
    ----------
    field : Field instance
        The field defining the reference mapping.
    integral : Integral instance
        The integral defining quadrature points.
    region : Region instance
        The given of the element faces.

    Returns
    -------
    normals : array
        The normals merged for all element groups.

    See Also
    --------
    get_mapping_data

    Notes
    -----
    Assumes the same element geometry in all element groups of the field!
    """
    normals = get_mapping_data('normal', field, integral, region=region,
                               integration='surface')
    return normals



r"""
Diametrically point loaded 2-D disk with postprocessing. See
:ref:`sec-primer`.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    = 0
    \;, \quad \forall \ul{v} \;,

where

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl}
    \;.
"""

from __future__ import absolute_import
from examples.linear_elasticity.its2D_1 import *

from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson

def stress_strain(out, pb, state, extend=False):
    """
    Calculate and output strain and stress for given displacements.
    """
    from sfepy.base.base import Struct

    ev = pb.evaluate
    strain = ev('ev_cauchy_strain.2.Omega(u)', mode='el_avg')
    stress = ev('ev_cauchy_stress.2.Omega(Asphalt.D, u)', mode='el_avg',
                copy_materials=False)

    out['cauchy_strain'] = Struct(name='output_data', mode='cell',
                                  data=strain, dofs=None)
    out['cauchy_stress'] = Struct(name='output_data', mode='cell',
                                  data=stress, dofs=None)

    return out

asphalt = materials['Asphalt'][0]
asphalt.update({'D' : stiffness_from_youngpoisson(2, young, poisson)})
options.update({'post_process_hook' : 'stress_strain',})



#!/usr/bin/env python

def configuration(parent_package='', top_path=None):
    import os.path as op
    from numpy.distutils.misc_util import Configuration

    from sfepy import Config

    site_config = Config()
    system = site_config.system()
    os_flag = {'posix' : 0, 'windows' : 1}[system]

    auto_dir = op.dirname(__file__)
    auto_name = op.split(auto_dir)[-1]
    config = Configuration(auto_name, parent_package, top_path)

    inline = 'inline' if system == 'posix' else '__inline'
    defines = [('SFEPY_PLATFORM', os_flag),
               ('inline', inline)]
    if '-DDEBUG_FMF' in site_config.debug_flags():
        defines.append(('DEBUG_FMF', None))
    if '-DDEBUG_MESH' in site_config.debug_flags():
        defines.append(('DEBUG_MESH', None))

    common_src = ['fmfield.c', 'refmaps.c', 'geommech.c', 'common_python.c']

    config.add_library('sfepy_common',
                       sources=common_src,
                       extra_compiler_args=site_config.compile_flags(),
                       extra_link_args=site_config.link_flags(),
                       include_dirs=[auto_dir, site_config.python_include()],
                       macros=[('SFEPY_PLATFORM', os_flag),
                               ('inline', inline)])

    src = ['_fmfield.pyx']
    config.add_extension('_fmfield',
                         sources=src,
                         libraries=['sfepy_common'],
                         depends=common_src,
                         extra_compile_args=site_config.compile_flags(),
                         extra_link_args=site_config.link_flags(),
                         include_dirs=[auto_dir],
                         define_macros=defines)

    src = ['mappings.pyx']
    config.add_extension('mappings',
                         sources=src,
                         libraries=['sfepy_common'],
                         depends=common_src,
                         extra_compile_args=site_config.compile_flags(),
                         extra_link_args=site_config.link_flags(),
                         include_dirs=[auto_dir],
                         define_macros=defines)

    src = ['assemble.pyx']
    config.add_extension('assemble',
                         sources=src,
                         extra_compile_args=site_config.compile_flags(),
                         extra_link_args=site_config.link_flags(),
                         include_dirs=[auto_dir],
                         define_macros=defines)

    src = ['cmesh.pyx', 'geomtrans.c', 'mesh.c', 'meshutils.c', 'sort.c',
           'common_python.c']
    config.add_extension('cmesh',
                         sources=src,
                         extra_compile_args=site_config.compile_flags(),
                         extra_link_args=site_config.link_flags(),
                         include_dirs=[auto_dir],
                         define_macros=defines)

    src = ['crefcoors.pyx', 'refcoors.c', 'geomtrans.c', 'mesh.c']
    config.add_extension('crefcoors',
                         sources=src,
                         libraries=['sfepy_common'],
                         depends=common_src,
                         extra_compile_args=site_config.compile_flags(),
                         extra_link_args=site_config.link_flags(),
                         include_dirs=[auto_dir],
                         define_macros=defines)

    src = ['_geommech.pyx']
    config.add_extension('_geommech',
                         sources=src,
                         libraries=['sfepy_common'],
                         extra_compile_args=site_config.compile_flags(),
                         extra_link_args=site_config.link_flags(),
                         include_dirs=[auto_dir],
                         define_macros=defines)

    # Include *.pxd files in distribution tarball and install them along
    # with the extension modules.
    pxd_files = ['cmesh.pxd', 'mappings.pxd', 'types.pxd',
                 '_fmfield.pxd', '_geommech.pxd']
    config.add_data_files(('', pxd_files))

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



from __future__ import absolute_import
import os

import numpy as nm

from sfepy.base.testing import TestCommon
from sfepy import data_dir
from six.moves import range

# n_vertex, n_edge, n_face, n_cell
# d1 -> d2 : num, n_incident
expected = {
    '1_2_2.mesh' : ([3, 2, 0, 0], {
        (0, 0) : (3, 4),
        (0, 1) : (3, 4),
        (1, 0) : (2, 4),
        (1, 1) : (2, 2),
        }),
    '2_3_2.mesh' : ([4, 5, 2, 0], {
        (0, 0) : (4, 10),
        (0, 1) : (4, 10),
        (0, 2) : (4, 6),
        (1, 0) : (5, 10),
        (1, 1) : (5, 16),
        (1, 2) : (5, 6),
        (2, 0) : (2, 6),
        (2, 1) : (2, 6),
        (2, 2) : (2, 2),
        }),
    '2_4_2.mesh' : ([6, 7, 2, 0], {
        (0, 0) : (6, 22),
        (0, 1) : (6, 14),
        (0, 2) : (6, 8),
        (1, 0) : (7, 14),
        (1, 1) : (7, 20),
        (1, 2) : (7, 8),
        (2, 0) : (2, 8),
        (2, 1) : (2, 8),
        (2, 2) : (2, 2),
        }),
    '3_4_2.mesh' : ([5, 9, 7, 2], {
        (0, 0) : (5, 18),
        (0, 1) : (5, 18),
        (0, 2) : (5, 21),
        (0, 3) : (5, 8),
        (1, 0) : (9, 18),
        (1, 1) : (9, 48),
        (1, 2) : (9, 21),
        (1, 3) : (9, 12),
        (2, 0) : (7, 21),
        (2, 1) : (7, 21),
        (2, 2) : (7, 42),
        (2, 3) : (7, 8),
        (3, 0) : (2, 8),
        (3, 1) : (2, 12),
        (3, 2) : (2, 8),
        (3, 3) : (2, 2),
        }),
    '3_8_2.mesh' : ([12, 20, 11, 2], {
        (0, 0) : (12, 100),
        (0, 1) : (12, 40),
        (0, 2) : (12, 44),
        (0, 3) : (12, 16),
        (1, 0) : (20, 40),
        (1, 1) : (20, 96),
        (1, 2) : (20, 44),
        (1, 3) : (20, 24),
        (2, 0) : (11, 44),
        (2, 1) : (11, 44),
        (2, 2) : (11, 72),
        (2, 3) : (11, 12),
        (3, 0) : (2, 16),
        (3, 1) : (2, 24),
        (3, 2) : (2, 12),
        (3, 3) : (2, 2),
        }),
    'square_triquad.mesh' : ([470, 1127, 658, 0], {
        (0, 0) : (470, 3054),
        (0, 1) : (470, 2254),
        (0, 2) : (470, 2174),
        (1, 0) : (1127, 2254),
        (1, 1) : (1127, 9174),
        (1, 2) : (1127, 2174),
        (2, 0) : (658, 2174),
        (2, 1) : (658, 2174),
        (2, 2) : (658, 6686),
        }),
}

class Test(TestCommon):

    @staticmethod
    def from_conf(conf, options):
        filename_meshes = [data_dir + '/meshes/elements/%s_2.mesh' % geom
                           for geom in ['1_2', '2_3', '2_4', '3_4', '3_8']]
        filename_meshes.append(data_dir
                               + '/meshes/2d/special/square_triquad.mesh')

        test = Test(filename_meshes=filename_meshes,
                    conf=conf, options=options)
        return test

    def test_cmesh_counts(self):
        from sfepy.discrete.fem import Mesh
        from sfepy.discrete.fem.geometry_element import create_geometry_elements
        from sfepy.discrete.common.extmods.cmesh import CMesh, get_cmem_usage

        gels = create_geometry_elements()

        ok = True

        for filename in self.filename_meshes:
            basename = os.path.basename(filename)
            enum, esizes = expected[basename]

            self.report('mesh: %s' % basename)

            mesh = Mesh.from_file(filename)
            cmesh = mesh.cmesh
            cmesh.set_local_entities(gels)

            cmesh.setup_entities()

            self.report('dim:', cmesh.dim)
            self.report('n_vertex: %d, n_edge: %d, n_face: %d, n_cell: %d' %
                        tuple(cmesh.num))

            _ok = (enum == cmesh.num).all()
            if not _ok:
                self.report('%s == %s failed!' % (enum, cmesh.num))
            ok = ok and _ok

            dim = cmesh.dim
            for ir in range(dim + 1):
                for ic in range(dim + 1):
                    cmesh.setup_connectivity(ir, ic)
                    mem_usage1 = get_cmem_usage()[0]

                    if (ir == dim) and (ic == 0):
                        continue

                    cmesh.free_connectivity(ir, ic)
                    mem_usage2 = get_cmem_usage()[0]

                    cmesh.setup_connectivity(ir, ic)
                    mem_usage3 = get_cmem_usage()[0]

                    conn = cmesh.get_conn(ir, ic)

                    self.report('(%d, %d) : (%d, %d)'
                                % (ir, ic, conn.num, conn.n_incident))
                    sizes = nm.array([conn.num, conn.n_incident])

                    _ok = (esizes[ir, ic] == sizes).all()
                    if not _ok:
                        self.report('%s == %s failed!' % (esizes, sizes))
                    ok = ok and _ok

                    _ok1 = mem_usage3 == mem_usage1
                    _ok2 = mem_usage3 > mem_usage2
                    if not (_ok1 and _ok2):
                        self.report('unexpected memory usage! (%s)'
                                    % (mem_usage1, mem_usage2, mem_usage3))
                    ok = ok and (_ok1 and _ok2)

        return ok

    def test_entity_volumes(self):
        import sfepy
        from sfepy.discrete.fem import Mesh, FEDomain
        from sfepy.discrete.common import Field
        from sfepy.discrete import Integral

        mesh = Mesh.from_file('meshes/3d/special/cross3d.mesh',
                              prefix_dir=sfepy.data_dir)
        domain = FEDomain('domain', mesh)

        omega = domain.create_region('Omega', 'all')
        gamma = domain.create_region('Gamma', 'vertices of surface', 'facet')
        top = domain.create_region('Top', 'cell 2')

        vfield = Field.from_args('v', nm.float64, 'scalar', omega,
                                 approx_order=1)
        sfield = Field.from_args('s', nm.float64, 'scalar', gamma,
                                 approx_order=1)

        integral = Integral('i', order=3)
        vgeo, _ = vfield.get_mapping(omega, integral, 'volume')
        domain.create_surface_group(gamma)
        sgeo, _ = sfield.get_mapping(gamma, integral, 'surface')

        evols = mesh.cmesh.get_volumes(1)
        fvols = mesh.cmesh.get_volumes(2) # Approximate for non-planar faces.
        cvols = mesh.cmesh.get_volumes(3)

        ok = True
        _ok = abs(cvols.sum() - vgeo.volume.sum()) < 1e-15
        self.report('total cell volume: %s (ok: %s)' % (cvols.sum(), _ok))
        ok = _ok and ok

        top_evols = nm.array([ 1.                ,  1.                ,
                               1.                ,  1.                ,
                               0.7211102550927979,  0.7211102550927979,
                               0.7211102550927979,  0.7211102550927979,
                               1.16619037896906  ,  1.16619037896906  ,
                               1.16619037896906  ,  1.16619037896906  ])

        _ok = nm.allclose(top_evols, evols[top.edges], rtol=0.0, atol=1e-15)
        self.report('total top cell edge length: %s (ok: %s)'
                    % (evols[top.edges].sum(), _ok))
        ok = _ok and ok

        i1 = [5, 6, 8, 9]
        i2 = nm.setdiff1d(nm.arange(len(gamma.faces)), i1)
        aux = fvols[gamma.faces] - sgeo.volume.ravel()

        _ok = nm.allclose(aux[i1], 0.10560208437556773, rtol=0.0, atol=1e-15)
        ok = _ok and ok
        self.report('non-planar faces diff: %s (ok: %s)' % (aux[i1], _ok))

        _ok = (nm.abs(aux[i2]) < 1e-15).all()
        self.report('max. planar faces diff: %s (ok: %s)'
                    % (nm.abs(aux[i2]).max(), _ok))
        ok = _ok and ok

        return ok



from __future__ import absolute_import
import os, glob

from .config import Config
from .version import __version__, in_source_tree, top_dir

data_dir = os.path.realpath(top_dir)
base_dir = os.path.dirname(os.path.normpath(os.path.realpath(__file__)))

def get_paths(pattern):
    """
    Get files/paths matching the given pattern in the sfepy source tree.
    """
    if not in_source_tree:
        pattern = '../' + pattern

    files = glob.glob(os.path.normpath(os.path.join(top_dir, pattern)))
    return files



r"""
Transient Laplace equation with non-constant initial conditions given by a
function.

Find :math:`T(t)` for :math:`t \in [0, t_{\rm final}]` such that:

.. math::
    \int_{\Omega} s \pdiff{T}{t}
    + \int_{\Omega} c \nabla s \cdot \nabla T
    = 0
    \;, \quad \forall s \;.
"""
from __future__ import absolute_import
from sfepy import data_dir

filename_mesh = data_dir + '/meshes/3d/cylinder.mesh'

t0 = 0.0
t1 = 0.1
n_step = 11

material_2 = {
    'name' : 'coef',
    'values' : {'val' : 0.01},
    'kind' : 'stationary', # 'stationary' or 'time-dependent'
}

field_1 = {
    'name' : 'temperature',
    'dtype' : 'real',
    'shape' : (1,),
    'region' : 'Omega',
    'approx_order' : 1,
}

variable_1 = {
    'name' : 'T',
    'kind' : 'unknown field',
    'field' : 'temperature',
    'order' : 0,
    'history' : 1,
}
variable_2 = {
    'name' : 's',
    'kind' : 'test field',
    'field' : 'temperature',
    'dual' : 'T',
}

regions = {
    'Omega' : 'all',
    'Gamma_Left' : ('vertices in (x < 0.00001)', 'facet'),
    'Gamma_Right' : ('vertices in (x > 0.099999)', 'facet'),
}

ebcs = {
    'T1': ('Gamma_Left', {'T.0' : 2.0}),
    'T2': ('Gamma_Right', {'T.0' : -2.0}),
}

def get_ic(coor, ic):
    """Non-constant initial condition."""
    import numpy as nm
    # Normalize x coordinate.
    mi, ma = coor[:,0].min(), coor[:,0].max()
    nx = (coor[:,0] - mi) / (ma - mi)
    return nm.where( (nx > 0.25) & (nx < 0.75 ), 8.0 * (nx - 0.5), 0.0 )

functions = {
    'get_ic' : (get_ic,),
}

ics = {
    'ic' : ('Omega', {'T.0' : 'get_ic'}),
}

integral_1 = {
    'name' : 'i',
    'order' : 1,
}

equations = {
    'Temperature' :
    """dw_volume_dot.i.Omega( s, dT/dt )
     + dw_laplace.i.Omega( coef.val, s, T ) = 0"""
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
    'use_presolve' : True,
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',

    'i_max'      : 1,
    'eps_a'      : 1e-10,
    'eps_r'      : 1.0,
    'macheps'   : 1e-16,
    'lin_red'    : 1e-2, # Linear system error < (eps_a * lin_red).
    'ls_red'     : 0.1,
    'ls_red_warp' : 0.001,
    'ls_on'      : 1.1,
    'ls_min'     : 1e-5,
    'check'     : 0,
    'delta'     : 1e-6,
    'is_linear' : True,
}

solver_2 = {
    'name' : 'ts',
    'kind' : 'ts.simple',

    't0'    : t0,
    't1'    : t1,
    'dt'    : None,
    'n_step' : n_step, # has precedence over dt!
    'verbose' : 1,
}

options = {
    'nls' : 'newton',
    'ls' : 'ls',
    'ts' : 'ts',
    'save_times' : 'all',
}



from __future__ import absolute_import
import os.path as op

from sfepy.base.testing import TestCommon

def get_volume(el, nd):
    from sfepy.mesh.mesh_tools import elems_q2t
    from sfepy.base.compat import factorial
    import numpy as nm
    from sfepy.linalg.utils import dets_fast

    dim = nd.shape[1]
    nnd = el.shape[1]

    etype = '%d_%d' % (dim, nnd)
    if etype == '2_4' or etype == '3_8':
        el = elems_q2t(el)

    nel = el.shape[0]

    mul = 1.0 / factorial(dim)
    if dim == 3:
        mul *= -1.0

    mtx = nm.ones((nel, dim + 1, dim + 1), dtype=nm.double)
    mtx[:,:,:-1] = nd[el,:]
    vols = mul * dets_fast(mtx)
    vol = vols.sum()

    return vol

class Test(TestCommon):

    @staticmethod
    def from_conf(conf, options):
        test = Test(conf=conf, options=options)
        return test

    def test_mesh_smoothing(self):
        from sfepy.mesh.mesh_tools import smooth_mesh
        from sfepy.discrete.fem.mesh import Mesh
        from sfepy import data_dir

        mesh = Mesh.from_file(data_dir + '/meshes/3d/cylinder.vtk')
        conn = mesh.get_conn('3_4')
        vol0 = get_volume(conn, mesh.coors)
        mesh.coors[:] = smooth_mesh(mesh, n_iter=10)
        vol1 = get_volume(conn, mesh.coors)
        filename = op.join(self.options.out_dir, 'smoothed_cylinder.vtk')
        mesh.write(filename)
        frac = vol1 / vol0

        if (frac < 0.967) and (frac > 0.966):
            self.report('mesh smoothed')
            return True

        else:
            self.report('mesh smoothed, volume mismatch!')
            return False



"""
Finite element reference mappings.
"""
import numpy as nm

from sfepy import Config
from sfepy.base.base import get_default, output
from sfepy.base.mem_usage import raise_if_too_large
from sfepy.discrete.common.mappings import Mapping
from sfepy.discrete.common.extmods.mappings import CMapping
from sfepy.discrete import PolySpace

class FEMapping(Mapping):
    """
    Base class for finite element mappings.
    """

    def __init__(self, coors, conn, poly_space=None, gel=None, order=1):
        self.coors = coors
        self.conn = conn

        try:
            nm.take(self.coors, self.conn)

        except IndexError:
            output('coordinates shape: %s' % list(coors.shape))
            output('connectivity: min: %d, max: %d' % (conn.min(), conn.max()))
            msg = 'incompatible connectivity and coordinates (see above)'
            raise IndexError(msg)

        self.n_el, self.n_ep = conn.shape
        self.dim = self.coors.shape[1]

        if poly_space is None:
            poly_space = PolySpace.any_from_args(None, gel, order,
                                                 base='lagrange',
                                                 force_bubble=False)

        self.poly_space = poly_space
        self.indices = slice(None)

    def get_geometry(self):
        """
        Return reference element geometry as a GeometryElement instance.
        """
        return self.poly_space.geometry

    def get_base(self, coors, diff=False):
        """
        Get base functions or their gradient evaluated in given
        coordinates.
        """
        bf = self.poly_space.eval_base(coors, diff=diff)
        return bf

    def get_physical_qps(self, qp_coors):
        """
        Get physical quadrature points corresponding to given reference
        element quadrature points.

        Returns
        -------
        qps : array
            The physical quadrature points ordered element by element,
            i.e. with shape (n_el, n_qp, dim).
        """
        bf = self.get_base(qp_coors)
        qps = nm.dot(nm.atleast_2d(bf.squeeze()), self.coors[self.conn])
        # Reorder so that qps are really element by element.
        qps = nm.ascontiguousarray(nm.swapaxes(qps, 0, 1))

        return qps

class VolumeMapping(FEMapping):
    """
    Mapping from reference domain to physical domain of the same space
    dimension.
    """

    def get_mapping(self, qp_coors, weights, poly_space=None, ori=None,
                    transform=None):
        """
        Get the mapping for given quadrature points, weights, and
        polynomial space.

        Returns
        -------
        cmap : CMapping instance
            The volume mapping.
        """
        poly_space = get_default(poly_space, self.poly_space)

        bf_g = self.get_base(qp_coors, diff=True)

        ebf_g = poly_space.eval_base(qp_coors, diff=True, ori=ori,
                                     force_axis=True, transform=transform)
        size = ebf_g.nbytes * self.n_el
        site_config = Config()
        raise_if_too_large(size, site_config.refmap_memory_factor())

        flag = (ori is not None) or (ebf_g.shape[0] > 1)

        cmap = CMapping(self.n_el, qp_coors.shape[0], self.dim,
                        poly_space.n_nod, mode='volume', flag=flag)
        cmap.describe(self.coors, self.conn, bf_g, ebf_g, weights)

        return cmap

class SurfaceMapping(FEMapping):
    """
    Mapping from reference domain to physical domain of the space
    dimension higher by one.
    """

    def set_basis_indices(self, indices):
        """
        Set indices to cell-based basis that give the facet-based basis.
        """
        self.indices = indices

    def get_base(self, coors, diff=False):
        """
        Get base functions or their gradient evaluated in given
        coordinates.
        """
        bf = self.poly_space.eval_base(coors, diff=diff)
        ii = max(self.dim - 1, 1)
        return nm.ascontiguousarray(bf[..., :ii:, self.indices])

    def get_mapping(self, qp_coors, weights, poly_space=None, mode='surface'):
        """
        Get the mapping for given quadrature points, weights, and
        polynomial space.

        Returns
        -------
        cmap : CMapping instance
            The surface mapping.
        """
        poly_space = get_default(poly_space, self.poly_space)

        bf_g = self.get_base(qp_coors, diff=True)
        if nm.allclose(bf_g, 0.0) and self.dim > 1:
            raise ValueError('zero base function gradient!')

        cmap = CMapping(self.n_el, qp_coors.shape[0], self.dim,
                        poly_space.n_nod, mode=mode)
        cmap.describe(self.coors, self.conn, bf_g, None, weights)
        if self.dim == 1:
            # Fix normals.
            ii = nm.where(self.conn == 0)[0]
            cmap.normal[ii] *= -1.0

        return cmap



#!/usr/bin/env python
from __future__ import print_function
from __future__ import absolute_import
from argparse import ArgumentParser
import numpy as nm

import sys
sys.path.append('.')

from sfepy.base.base import IndexedStruct
from sfepy.discrete import (FieldVariable, Material, Integral, Function,
                            Equation, Equations, Problem)
from sfepy.discrete.fem import Mesh, FEDomain, Field
from sfepy.terms import Term
from sfepy.discrete.conditions import Conditions, EssentialBC
from sfepy.solvers.ls import ScipyDirect
from sfepy.solvers.nls import Newton
from sfepy.postprocess.viewer import Viewer
from sfepy.mechanics.matcoefs import stiffness_from_lame


def shift_u_fun(ts, coors, bc=None, problem=None, shift=0.0):
    """
    Define a displacement depending on the y coordinate.
    """
    val = shift * coors[:,1]**2

    return val

helps = {
    'show' : 'show the results figure',
}

def main():
    from sfepy import data_dir

    parser = ArgumentParser()
    parser.add_argument('--version', action='version', version='%(prog)s')
    parser.add_argument('-s', '--show',
                        action="store_true", dest='show',
                        default=False, help=helps['show'])
    options = parser.parse_args()

    mesh = Mesh.from_file(data_dir + '/meshes/2d/rectangle_tri.mesh')
    domain = FEDomain('domain', mesh)

    min_x, max_x = domain.get_mesh_bounding_box()[:,0]
    eps = 1e-8 * (max_x - min_x)
    omega = domain.create_region('Omega', 'all')
    gamma1 = domain.create_region('Gamma1',
                                  'vertices in x < %.10f' % (min_x + eps),
                                  'facet')
    gamma2 = domain.create_region('Gamma2',
                                  'vertices in x > %.10f' % (max_x - eps),
                                  'facet')

    field = Field.from_args('fu', nm.float64, 'vector', omega,
                            approx_order=2)

    u = FieldVariable('u', 'unknown', field)
    v = FieldVariable('v', 'test', field, primary_var_name='u')

    m = Material('m', D=stiffness_from_lame(dim=2, lam=1.0, mu=1.0))
    f = Material('f', val=[[0.02], [0.01]])

    integral = Integral('i', order=3)

    t1 = Term.new('dw_lin_elastic(m.D, v, u)',
                  integral, omega, m=m, v=v, u=u)
    t2 = Term.new('dw_volume_lvf(f.val, v)', integral, omega, f=f, v=v)
    eq = Equation('balance', t1 + t2)
    eqs = Equations([eq])

    fix_u = EssentialBC('fix_u', gamma1, {'u.all' : 0.0})

    bc_fun = Function('shift_u_fun', shift_u_fun,
                      extra_args={'shift' : 0.01})
    shift_u = EssentialBC('shift_u', gamma2, {'u.0' : bc_fun})

    ls = ScipyDirect({})

    nls_status = IndexedStruct()
    nls = Newton({}, lin_solver=ls, status=nls_status)

    pb = Problem('elasticity', equations=eqs)
    pb.save_regions_as_groups('regions')

    pb.set_bcs(ebcs=Conditions([fix_u, shift_u]))

    pb.set_solver(nls)

    status = IndexedStruct()
    state = pb.solve(status=status)

    print('Nonlinear solver status:\n', nls_status)
    print('Stationary solver status:\n', status)

    pb.save_state('linear_elasticity.vtk', state)

    if options.show:
        view = Viewer('linear_elasticity.vtk')
        view(vector_mode='warp_norm', rel_scaling=2,
             is_scalar_bar=True, is_wireframe=True)

if __name__ == '__main__':
    main()



"""
Helper functions related to mesh facets and Lagrange FE approximation.

Line: ori - iter:

0 - iter0
1 - iter1

Triangle: ori - iter:

0 - iter21
1 - iter12
3 - iter02
4 - iter20
6 - iter10
7 - iter01

Possible couples:

1, 4, 7 <-> 0, 3, 6

Square: ori - iter:

 0 - iter10x01y
 7 - iter10y01x
11 - iter01y01x
30 - iter01x10y
33 - iter10x10y
52 - iter01y10x
56 - iter10y10x
63 - iter01x01y

Possible couples:

7, 33, 52, 63 <-> 0, 11, 30, 56
"""
from __future__ import print_function
from __future__ import absolute_import
import numpy as nm
import six
from six.moves import range

_quad_ori_groups = {
    0 : 0,
    1 : 0,
    3 : 7,
    4 : 0,
    6 : 7,
    7 : 7,
    11 : 11,
    15 : 11,
    20 : 52,
    22 : 30,
    30 : 30,
    31 : 63,
    32 : 33,
    33 : 33,
    41 : 33,
    43 : 11,
    48 : 56,
    52 : 52,
    56 : 56,
    57 : 56,
    59 : 63,
    60 : 52,
    62 : 30,
    63 : 63,
}

def build_orientation_map(n_fp):
    """
    The keys are binary masks of the lexicographical ordering of facet
    vertices. A bit i set to one means `v[i] < v[i+1]`.

    The values are `[original_order, permutation]`, where `permutation` can be
    used to sort facet vertices lexicographically. Hence `permuted_facet =
    facet[permutation]`.
    """
    from sfepy.linalg import permutations

    indices = list(range(n_fp))

    cmps = [(i1, i2) for i2 in indices for i1 in indices[:i2]]
    powers = [2**ii for ii in range(len(cmps))]

    ori_map = {}
    for indx in permutations(indices):
        key = 0
        sign = 1
        for ip, power in enumerate(powers):
            i1, i2 = cmps[ip]
            less = (indx[i1] < indx[i2])
            key += power * less
            if not less:
                sign *= -1

        isort = nm.argsort(indx)
        ori_map[key] = [indx, isort]

    return ori_map, cmps, powers

def iter0(num):
    for ir in range(num - 1, -1, -1):
        yield ir

def iter1(num):
    for ir in range(num):
        yield ir

ori_line_to_iter = {
    0 : iter0,
    1 : iter1,
}

def make_line_matrix(order):
    if (order < 2):
        return nm.zeros((0, 0), dtype=nm.int32)

    oo = order - 1
    mtx = nm.arange(oo, dtype=nm.int32)

    return mtx

def iter01(num):
    for ir in range(num - 1, -1, -1):
        for ic in range(ir + 1):
            yield ir, ic

def iter10(num):
    for ir in range(num - 1, -1, -1):
        for ic in range(ir, -1, -1):
            yield ir, ic

def iter02(num):
    for ic in range(num):
        for ir in range(num - 1, ic - 1, -1):
            yield ir, ic

def iter20(num):
    for ic in range(num):
        for ir in range(ic, num):
            yield ir, ic

def iter12(num):
    for idiag in range(num):
        irs, ics = nm.diag_indices(num - idiag)
        for ii in range(irs.shape[0] - 1, -1, -1):
            yield irs[ii] + idiag, ics[ii]

def iter21(num):
    for idiag in range(num):
        irs, ics = nm.diag_indices(num - idiag)
        for ii in range(irs.shape[0]):
            yield irs[ii] + idiag, ics[ii]

ori_triangle_to_iter = {
    0 : iter21,
    1 : iter12,
    3 : iter02,
    4 : iter20,
    6 : iter10,
    7 : iter01,
}

def make_triangle_matrix(order):
    if (order < 3):
        return nm.zeros((0, 0), dtype=nm.int32)

    oo = order - 2
    mtx = nm.zeros((oo, oo), dtype=nm.int32)
    for ii, (ir, ic) in enumerate(iter01(oo)):
        mtx[ir, ic] = ii

    return mtx

def iter01x01y(num):
    for ir in range(num):
        for ic in range(num):
            yield ir, ic

def iter01y01x(num):
    for ir, ic in iter01x01y(num):
        yield ic, ir

def iter10x01y(num):
    for ir in range(num - 1, -1, -1):
        for ic in range(num):
            yield ir, ic

def iter10y01x(num):
    for ir, ic in iter10x01y(num):
        yield ic, ir

def iter01x10y(num):
    for ir in range(num):
        for ic in range(num - 1, -1, -1):
            yield ir, ic

def iter01y10x(num):
    for ir, ic in iter01x10y(num):
        yield ic, ir

def iter10x10y(num):
    for ir in range(num - 1, -1, -1):
        for ic in range(num - 1, -1, -1):
            yield ir, ic

def iter10y10x(num):
    for ir, ic in iter10x10y(num):
        yield ic, ir

ori_square_to_iter = {
    0 : iter10x01y,
    7 : iter10y01x,
    11 : iter01y01x,
    30 : iter01x10y,
    33 : iter10x10y,
    52 : iter01y10x,
    56 : iter10y10x,
    63 : iter01x01y,
}

def make_square_matrix(order):
    if (order < 2):
        return nm.zeros((0, 0), dtype=nm.int32)

    oo = order - 1
    mtx = nm.arange(oo * oo, dtype=nm.int32)
    mtx.shape = (oo, oo)

    return mtx

def get_facet_dof_permutations(n_fp, order):
    """
    Prepare DOF permutation vector for each possible facet orientation.
    """
    from sfepy.base.base import dict_to_array

    if n_fp == 2:
        mtx = make_line_matrix(order)
        ori_map = ori_line_to_iter
        fo = order - 1

    elif n_fp == 3:
        mtx = make_triangle_matrix(order)
        ori_map = ori_triangle_to_iter
        fo = order - 2

    elif n_fp == 4:
        mtx = make_square_matrix(order)
        ori_map = {}
        for key, val in six.iteritems(_quad_ori_groups):
            ori_map[key] = ori_square_to_iter[val]
        fo = order - 1

    else:
        raise ValueError('unsupported number of facet points! (%d)' % n_fp)

    dof_perms = {}
    for key, itfun in six.iteritems(ori_map):
        dof_perms[key] = [mtx[ii] for ii in itfun(fo)]

    dof_perms = dict_to_array(dof_perms)

    return dof_perms

if __name__ == '__main__':
    order = 5
    mtx = make_triangle_matrix(order)
    print(mtx)

    oo = order - 2
    print([mtx[ir, ic] for ir, ic in ori_triangle_to_iter[0](oo)])
    print([mtx[ir, ic] for ir, ic in ori_triangle_to_iter[1](oo)])
    print([mtx[ir, ic] for ir, ic in ori_triangle_to_iter[3](oo)])
    print([mtx[ir, ic] for ir, ic in ori_triangle_to_iter[4](oo)])
    print([mtx[ir, ic] for ir, ic in ori_triangle_to_iter[6](oo)])
    print([mtx[ir, ic] for ir, ic in ori_triangle_to_iter[7](oo)])

    order = 4
    mtx = make_square_matrix(order)
    print(mtx)

    oo = order - 1
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[0](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[7](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[11](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[30](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[33](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[52](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[56](oo)])
    print([mtx[ir, ic] for ir, ic in ori_square_to_iter[63](oo)])



"""
Functions for resolving dependencies.
"""
from __future__ import absolute_import
import itertools as it

from sfepy.base.base import basestr
import six
from six.moves import range

def get_nums(deps):
    """
    Get number of prerequisite names for each name in dependencies.
    """
    nums = {}
    for key, val in six.iteritems(deps):
        nums[key] = len(val)

    return nums

def solvable(deps, names):
    """
    Return True if `names` form a solvable block, i.e. the set of names equals
    to the set of their prerequisites.
    """
    if not names: return False # Missing self-reference.

    dep_names = set()
    for name in names:
        dep_names.update(deps[name])

    return dep_names == set(names)

def remove_known(deps, known):
    """
    Remove known names from dependencies.
    """
    if isinstance(known, basestr):
        out = {}
        for key, val in six.iteritems(deps):
            if key == known: continue
            out[key] = [ii for ii in val if ii != known]

        return out

    else:
        out = deps
        for ii in known:
            out = remove_known(out, ii)

        return out

def try_block(deps, num):
    """
    Return generator of lists of solvable blocks of the length `num`.
    """
    keys = list(deps.keys())
    for ic in it.combinations(keys, num):
        if solvable(deps, ic):
            yield sorted(ic)

def resolve(deps):
    """
    Resolve dependencies among equations so that smaller blocks are solved
    first.

    The dependencies are given in terms of variable names.

    Parameters
    ----------
    deps : dict
        The dependencies as a dictionary with names as keys and sets of
        prerequisite names as values.

    Returns
    -------
    order : list
        The list of blocks in the order of solving. Each block is a list of
        names.
    """
    order = []
    if not(len(deps)): return order

    nums = get_nums(deps)
    ib0 = min(nums.values())
    for ib in range(ib0, len(deps) + 1):
        blocks = [ii for ii in try_block(deps, ib)]
        if len(blocks):
            new_deps = remove_known(deps, blocks[0])
            order.extend([blocks[0]] + resolve(new_deps))

        if len(list(it.chain(*order))) == len(deps):
            break

    return order



# SfePy version
__version__ = '2020.3'

# "Minimal" supported versions.
NUMPY_MIN_VERSION = '1.3'
SCIPY_MIN_VERSION = '0.7'
MATPLOTLIB_MIN_VERSION = '0.99.0'
PYPARSING_MIN_VERSION = '1.5.0'
PYTABLES_MIN_VERSION = '2.1.2'
MAYAVI_MIN_VERSION = '3.3.0'
SYMPY_MIN_VERSION = '0.7.3'
IGAKIT_MIN_VERSION = '0.1'
PETSC4PY_MIN_VERSION = '3.4'
SLEPC4PY_MIN_VERSION = '3.4'
MPI4PY_MIN_VERSION = '1.3.1'
PYMETIS_MIN_VERSION = '2014.1'
SCIKIT_UMFPACK_MIN_VERSION = '0.1'
MESHIO_MIN_VERSION='4.0.0'
PSUTIL_MIN_VERSION='1.0.0'
PYVISTA_MIN_VERSION='0.23.0'

CYTHON_MIN_VERSION = '0.14.1'

def get_basic_info(version=__version__):
    """
    Return SfePy installation directory information. Append current git
    commit hash to `version`.
    """
    import os.path as op
    from sfepy import Config

    # If installed, up_dir is '.', otherwise (in (git) source directory) '..'.
    for up_dir in ['..', '.']:
        top_dir = op.normpath(op.realpath(op.join(op.dirname(__file__),
                                                  up_dir)))
        aux = op.join(top_dir, 'README.rst')
        if op.isfile(aux):
            break
    else:
        raise RuntimeError('cannot determine SfePy top level directory!')

    config = Config()
    if not config.is_release():
        # Append current git commit hash to __version__.
        master = op.join(top_dir, '.git/refs/heads/master')
        if op.isfile(master):
            fd = open(master, 'r')
            version += '+git.%s' % fd.readline().strip()
            fd.close()

    in_source_tree = up_dir == '..'

    return version, top_dir, in_source_tree

__version__, top_dir, in_source_tree = get_basic_info(__version__)



"""
Functions to visualize the partitioning of a domain and a field DOFs.
"""
import os

import numpy as nm

from sfepy.base.base import ordered_iteritems
import sfepy.postprocess.plot_cmesh as pc
import sfepy.postprocess.plot_dofs as pd

def mark_subdomains(ax, cmesh, cell_tasks,
                    size=None, icolor=0, alpha=1.0, mask=False):
    """
    Mark cells of subdomains corresponding to each task by a different color.
    Plots nothing in 3D.
    """
    if size is None:
        size = cell_tasks.max() + 1

    coors = cmesh.coors
    dim = cmesh.dim

    ax = pd._get_axes(ax, dim)

    if dim == 3:
        return ax

    conn = cmesh.get_conn(dim, 0)

    color = nm.zeros(4)
    color[-1] = alpha
    for ic, vertices in enumerate(conn.indices.reshape((conn.num, -1))):
        cv = coors[vertices]
        if dim == 2:
            if mask:
                cc = cv.mean(0)
                cv = cc + 0.3 * (cv - cc)

            if not mask or cell_tasks[ic] > 0:
                color[icolor] = (float(cell_tasks[ic]) + 1) / size
                ax.fill(*cv.T, color=color)

    return ax

def label_dofs(ax, coors, dofs, colors):
    """
    Label DOFs using the given colors.
    """
    from sfepy.postprocess.plot_dofs import _get_axes

    dim = coors.shape[1]
    ax = _get_axes(ax, dim)

    for gdof in dofs:
        cd = coors[gdof]
        ax.text(*cd.T, s='%d' % gdof,
                color=colors[gdof], fontsize=12, weight='bold')

    return ax

def plot_partitioning(axs, field, cell_tasks, gfd, output_dir, size):
    """
    Plot the partitioning of the domain and field DOFs.
    """
    mesh = field.domain.mesh

    ax = pc.plot_wireframe(axs[0], mesh.cmesh)
    coors = field.get_coor()
    econn = field.econn
    ax = pd.plot_global_dofs(ax, coors, econn)
    ax.set_title('global DOFs')
    ax.figure.savefig(os.path.join(output_dir, 'global_dofs.png'),
                      bbox_inches='tight')

    ax = pc.plot_wireframe(axs[1], mesh.cmesh)
    fig = ax.figure
    coors = field.get_coor()
    econn = field.econn

    id_map = gfd.id_map

    colors = nm.zeros((field.n_nod, 4))
    for ir, dof_map in ordered_iteritems(gfd.dof_maps):
        aux = id_map[dof_map[0]]
        colors[aux] = [0, 0, float(ir + 1) / size, 0.6]
        for aux in dof_map[1]:
            colors[id_map[aux]] = [0, 0, float(ir + 1) / size, 0.9]

    from sfepy.discrete.fem.utils import prepare_translate
    aux = prepare_translate(id_map[econn], econn)
    ax = label_dofs(ax, coors[aux], id_map, colors)

    mark_subdomains(ax, mesh.cmesh, cell_tasks, size, 0, 0.7)

    ax.set_title('subdomains of tasks and PETSc DOFs')

    fig.savefig(os.path.join(output_dir, 'petsc_dofs.png'),
                bbox_inches='tight')

    ax.set_title('')

    axis = ax.axis()
    for ir, ocells in enumerate(gfd.overlap_cells):
        aux = nm.zeros_like(cell_tasks)
        aux[ocells] = 10
        aux[gfd.cell_parts[ir]] = 1

        ax = fig.add_axes(ax.get_position(), frameon=False, label='aux')
        mark_subdomains(ax, mesh.cmesh, aux, 11, 1, 0.3, True)
        ax.axis(axis)
        ax.set_title('overlap cells on task %d' % ir)
        fig.savefig(os.path.join(output_dir,
                                 'petsc_overlaps_%02d.png' % ir),
                    bbox_inches='tight')
        fig.delaxes(ax)

def plot_local_dofs(axs, field, field_i, omega_gi, output_dir, rank):
    """
    Plot the local ang global field DOFs local to the subdomain on the task
    with the given `rank`.
    """
    mesh = field.domain.mesh

    ax = pc.plot_wireframe(axs[0], mesh.cmesh)
    coors = field_i.get_coor()
    econn = field_i.econn
    ax = pd.plot_global_dofs(ax, coors, econn)
    ax.set_title('local DOFs on task %d' % rank)
    ax.figure.savefig(os.path.join(output_dir, 'local_dofs_%02d.png' % rank),
                      bbox_inches='tight')

    ax = pc.plot_wireframe(axs[1], mesh.cmesh)
    coors = field.get_coor()
    econn = field.get_econn('volume', omega_gi, 0)
    ax = pd.plot_global_dofs(ax, coors, econn)
    ax.set_title('global DOFs on task %d' % rank)
    ax.figure.savefig(os.path.join(output_dir, 'local_global_%02d.png' % rank),
                      bbox_inches='tight')



#!/usr/bin/env python
# 05.10.2005, c
"""
Given a mesh file, this script extracts its surface and prints it to stdout in
form of a list where each row is [element, face, component]. A component
corresponds to a contiguous surface region - for example, a cubical mesh with a
spherical hole has two surface components. Two surface faces sharing a single
node belong to one component.

With '-m' option, a mesh of the surface is created and saved in
'<original path>/surf_<original mesh file name>.mesh'.
"""
from __future__ import absolute_import
import sys
from six.moves import range
sys.path.append('.')
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import numpy as nm
import scipy.sparse as sp

import sfepy
from sfepy.base.base import output
from sfepy.base.ioutils import edit_filename
from sfepy.discrete.fem import Mesh, FEDomain
from sfepy.discrete.common.extmods.cmesh import (create_mesh_graph,
                                                 graph_components)

def _get_facets(vertices, offsets, ii, n_fp):
    facets = []
    for ic in range(n_fp):
        facets.append(vertices[offsets[ii] + ic][:, None])

    facets = nm.concatenate(facets, axis=1)

    return nm.ascontiguousarray(facets.astype(nm.int32))

def get_surface_faces(domain):
    cmesh = domain.cmesh
    faces = cmesh.get_surface_facets()
    vertices_f, offs_f = cmesh.get_incident(0, faces,
                                            cmesh.dim - 1, ret_offsets=True)

    n_fp = nm.diff(offs_f)
    surf_faces = []

    itri = nm.where(n_fp == 3)[0]
    if itri.size:
        surf_faces.append(_get_facets(vertices_f, offs_f, itri, 3))

    itet = nm.where(n_fp == 4)[0]
    if itet.size:
        surf_faces.append(_get_facets(vertices_f, offs_f, itet, 4))

    cells_c, offs_c = cmesh.get_incident(cmesh.dim, faces, cmesh.dim - 1,
                                         ret_offsets=True)
    ids = cmesh.get_local_ids(faces, cmesh.dim - 1, cells_c, offs_c,
                              cmesh.dim)
    lst = nm.c_[cells_c, ids]

    return lst, surf_faces

def surface_graph(surf_faces, n_nod):
    nnz, prow, icol = create_mesh_graph(n_nod, n_nod, len(surf_faces),
                                        surf_faces, surf_faces)
    data = nm.empty((nnz,), dtype=nm.int32)
    data.fill(2)
    return sp.csr_matrix((data, icol, prow), (n_nod, n_nod))

def surface_components(gr_s, surf_faces):
    """
    Determine surface components given surface mesh connectivity graph.
    """
    n_nod = gr_s.shape[0]
    n_comp, flag = graph_components(n_nod, gr_s.indptr, gr_s.indices)

    comps = []
    for ii, face in enumerate(surf_faces):
        comp = flag[face[:,0]]
        comps.append(comp)

    return n_comp, comps

def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("--version", action="version",
                        version="%(prog)s " + sfepy.__version__)
    parser.add_argument("-m", "--mesh",
                        action="store_true", dest="save_mesh",
                        default=False,
                        help="save surface mesh")
    parser.add_argument("-n", "--no-surface",
                        action="store_true", dest="no_surface",
                        default=False,
                        help="do not output surface [default: %(default)s]")
    parser.add_argument('filename_in', help="'-' is for stdin")
    parser.add_argument('filename_out', help="'-' is for  stdout")
    options = parser.parse_args()

    filename_in = options.filename_in
    filename_out = options.filename_out

    if (filename_in == '-'):
        file_in = sys.stdin
    else:
        file_in = open(filename_in, "r")

    mesh = Mesh.from_file(filename_in)

    if (filename_in != '-'):
        file_in.close()

    domain = FEDomain('domain', mesh)

    if options.save_mesh:
        region = domain.create_region('surf', 'vertices of surface', 'facet')
        surf_mesh = Mesh.from_region(region, mesh,
                                     localize=True, is_surface=True)
        aux = edit_filename(filename_in, prefix='surf_', new_ext='.mesh')
        surf_mesh.write(aux, io='auto')

    if domain.has_faces():
        domain.fix_element_orientation()

        lst, surf_faces = get_surface_faces(domain)

        if options.no_surface:
            return

        gr_s = surface_graph(surf_faces, mesh.n_nod)

        n_comp, comps = surface_components(gr_s, surf_faces)
        output('number of surface components:', n_comp)

        ccs, comps = comps, nm.zeros((0,1), nm.int32)
        for cc in ccs:
            comps = nm.concatenate((comps, cc[:,nm.newaxis]), 0)

        out = nm.concatenate((lst, comps), 1)

        if (filename_out == '-'):
            file_out = sys.stdout
        else:
            file_out = open(filename_out, "w")
        for row in out:
            file_out.write('%d %d %d\n' % (row[0], row[1], row[2]))
        if (filename_out != '-'):
            file_out.close()

if __name__=='__main__':
    main()



#!/usr/bin/env python

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    import os.path as op

    auto_name = op.split(op.dirname(__file__))[-1]
    config = Configuration(auto_name, parent_package, top_path)

#    config.add_data_dir('tests')

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



r"""
This example shows how to use the post_process_hook to probe the output data.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    = 0
    \;, \quad \forall \ul{v} \;,

where

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl}
    \;.
"""
from __future__ import absolute_import
from examples.linear_elasticity.linear_elastic import *
import os
import numpy as nm

from sfepy.base.base import Struct
try:
    from sfepy.postprocess.probes_vtk import Probe

except ImportError:
    Probe = None

from six.moves import range

# Define options.
options = {
    'output_dir' : '.',
    'post_process_hook' : 'post_process',
}

# The function returning the probe parametrization.
def par_fun(idx):
    return nm.log(idx + 1) / nm.log(20) * 0.04

# Define the function post_process, that will be called after the problem is
# solved.
def post_process(out, problem, state, extend=False):
    """
    This will be called after the problem is solved.

    Parameters
    ----------
    out : dict
        The output dictionary, where this function will store additional data.
    problem : Problem instance
        The current Problem instance.
    state : State instance
        The computed state, containing FE coefficients of all the unknown
        variables.
    extend : bool
        The flag indicating whether to extend the output data to the whole
        domain. It can be ignored if the problem is solved on the whole domain
        already.

    Returns
    -------
    out : dict
        The updated output dictionary.
    """
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.font_manager as fm

    # Cauchy strain averaged in elements.
    strain = problem.evaluate('ev_cauchy_strain.i.Omega( u )',
                              mode='el_avg')
    out['cauchy_strain'] = Struct(name='output_data',
                                  mode='cell', data=strain,
                                  dofs=None)
    # Cauchy stress averaged in elements.
    stress = problem.evaluate('ev_cauchy_stress.i.Omega( solid.D, u )',
                              mode='el_avg')
    out['cauchy_stress'] = Struct(name='output_data',
                                  mode='cell', data=stress,
                                  dofs=None)

    if Probe is None:
        # Do not probe if vtk cannot be imported.
        return out

    # Define three line probes in axial directions.
    mesh = problem.domain.mesh

    bbox = mesh.get_bounding_box()
    cx, cy, cz = 0.5 * bbox.sum(axis=0)

    labels = []
    probe_names = []

    probe = Probe(out, mesh)

    # line probe
    labels.append('line probe - x direction')
    probe_names.append('line')
    probe.add_line_probe('line',
                         [bbox[0,0], cy, cz],
                         [bbox[1,0], cy, cz],
                         30)

    # circle probe
    labels.append('circle probe')
    probe_names.append('circle')
    probe.add_circle_probe('circle',
                           [cx, cy, cz],
                           [0, 0, 1],
                           0.015,
                           30)

    # ray probe
    labels.append('ray probe - y direction')
    probe_names.append('ray')
    probe.add_ray_probe('ray',
                        [cx, bbox[0,1], cz],
                        [0, 1, 0],
                        par_fun, 20)

    # Gnerate matplotlib figures with the probe plot.
    for probe_name, label in zip(probe_names, labels):

        fig = plt.figure()
        plt.clf()
        fig.subplots_adjust(hspace=0.4)

        plt.subplot(311)
        pars, vals = probe(probe_name, 'u')
        for ic in range(vals.shape[1]):
            plt.plot(pars, vals[:,ic], label=r'$u_{%d}$' % (ic + 1),
                     lw=1, ls='-', marker='+', ms=3)
        plt.ylabel('displacements')
        plt.xlabel('probe %s' % label, fontsize=8)
        plt.legend(loc='best', prop=fm.FontProperties(size=10))

        sym_labels = ['11', '22', '33', '12', '13', '23']

        plt.subplot(312)
        pars, vals = probe(probe_name, 'cauchy_strain')
        for ii, lab in enumerate(sym_labels):
            plt.plot(pars, vals[:,ii], label=r'$e_{%s}$' % lab,
                     lw=1, ls='-', marker='+', ms=3)
        plt.ylabel('Cauchy strain')
        plt.xlabel('probe %s' % label, fontsize=8)
        plt.legend(loc='best', prop=fm.FontProperties(size=8))

        plt.subplot(313)
        pars, vals = probe(probe_name, 'cauchy_stress')
        for ii, lab in enumerate(sym_labels):
            plt.plot(pars, vals[:,ii], label=r'$\tau_{%s}$' % lab,
                     lw=1, ls='-', marker='+', ms=3)
        plt.ylabel('Cauchy stress')
        plt.xlabel('probe %s' % label, fontsize=8)
        plt.legend(loc='best', prop=fm.FontProperties(size=8))

        opts = problem.conf.options
        filename_results = os.path.join(opts.get('output_dir'),
                                        'cylinder_probe_%s.png' % probe_name)

        fig.savefig(filename_results)

    return out



##
# Template file for site configuration - copy it to site_cfg.py:
# $ cp site_cfg_template.py site_cfg.py

# Set python version (e.g. '2.7', '3.7'}), or leave '3.*' for autodetection.
python_version = '3.*'

# Operating system - one of 'posix', 'windows' or None for automatic
# determination.
system = None

# Extra flags added to the flags supplied by distutils to compile C
# extension modules.
compile_flags = '-g -O2'

# Extra flags added to the flags supplied by distutils to link C
# extension modules.
link_flags = ''

# Can be '' or one or several from '-DDEBUG_FMF', '-DDEBUG_MESH'. For
# developers internal use only.
debug_flags = ''

# Sphinx documentation uses numpydoc extension. Set the path here in case it is
# not installed in a standard location.
numpydoc_path = None

# True for a release, False otherwise. If False, current git commit hash
# is appended to version string, if the sources are in a repository.
is_release = False

# Tetgen executable path.
tetgen_path = '/usr/bin/tetgen'

# Reference mapping memory allocation safety factor (float), requires psutil
# installed. Use None for skipping the memory check.
refmap_memory_factor = None



import numpy as nm
from sfepy.terms.terms import Term, terms
from sfepy.base.base import Struct
import six

_msg_missing_data = 'missing family data!'

class HyperElasticFamilyData(Struct):
    """
    Base class for hyperelastic family data.

    The common (family) data are cached in the evaluate cache of state
    variable.
    """
    data_shapes = {
            'mtx_f': ('n_el', 'n_qp', 'dim', 'dim'),
            'det_f': ('n_el', 'n_qp', 1, 1),
            'sym_b': ('n_el', 'n_qp', 'sym', 1),
            'tr_b': ('n_el', 'n_qp', 1, 1),
            'in2_b': ('n_el', 'n_qp', 1, 1),
            'sym_c' : ('n_el', 'n_qp', 'sym', 1),
            'tr_c' : ('n_el', 'n_qp', 1, 1),
            'in2_c' : ('n_el', 'n_qp', 1, 1),
            'sym_inv_c' : ('n_el', 'n_qp', 'sym', 1),
            'green_strain': ('n_el', 'n_qp', 'sym', 1),
            'inv_f': ('n_el', 'n_qp', 'dim', 'dim'),
    }

    def init_data_struct(self, state_shape, name='family_data'):
        from sfepy.mechanics.tensors import dim2sym

        n_el, n_qp, dim, n_en, n_c = state_shape
        sym = dim2sym(dim); sym

        shdict = dict(( (k, v) for k, v in six.iteritems(locals())\
            if k in ['n_el', 'n_qp', 'dim', 'n_en', 'n_c', 'sym']))

        data = Struct(name=name)
        setattr(data, 'names', self.data_names)
        for key in self.data_names:
            shape = [shdict[sh] if type(sh) is str else sh\
                 for sh in self.data_shapes[key]]
            setattr(data, key, nm.zeros(shape, dtype=nm.float64))

        return data

    def __call__(self, state, region, integral, integration,
                 step=0, derivative=None):
        step_cache = state.evaluate_cache.setdefault(self.cache_name, {})
        cache = step_cache.setdefault(step, {})

        key = (region.name, integral.order, integration)
        data_key = key + (derivative,)

        if data_key in cache:
            data = cache[data_key]

        else:
            vg, _ = state.field.get_mapping(region,
                                            integral, integration,
                                            get_saved=True)

            vec = state(step=step, derivative=derivative)

            st_shape = state.get_data_shape(integral, integration, region.name)
            data = self.init_data_struct(st_shape)

            fargs = tuple([getattr(data, k) for k in self.data_names])
            fargs = fargs + (vec, vg, state.field.econn)

            self.family_function(*fargs)
            cache[data_key] = data

        return data

class HyperElasticBase(Term):
    """
    Base class for all hyperelastic terms in TL/UL formulation.

    `HyperElasticBase.__call__()` computes element contributions given either
    stress (-> residual) or tangent modulus (-> tangent sitffnes matrix),
    i.e. constitutive relation type (CRT) related data. The CRT data are
    computed in subclasses implementing particular CRT (e.g. neo-Hookean
    material), in self.compute_crt_data().

    Modes:

      - 0: total formulation
      - 1: updated formulation

    Notes
    -----
    This is not a proper Term!
    """
    arg_types = ('material', 'virtual', 'state')
    arg_shapes = {'material' : '1, 1', 'virtual' : ('D', 'state'),
                  'state' : 'D'}

    @staticmethod
    def integrate(out, val_qp, vg, fmode):
        if fmode == 2:
            out[:] = val_qp
            status = 0

        else:
            status = vg.integrate(out, val_qp, fmode)

        return status

    @staticmethod
    def function(out, fun, *args):
        return fun(out, *args)

    def __init__(self, *args, **kwargs):
        Term.__init__(self, *args, **kwargs)

        self.stress_cache = None

    def compute_stress(self, mat, family_data, **kwargs):
        out = nm.empty_like(family_data.green_strain)

        get = family_data.get
        fargs = [get(name, msg_if_none=_msg_missing_data)
                 for name in self.family_data_names]

        self.stress_function(out, mat, *fargs, **kwargs)

        return out

    def compute_tan_mod(self, mat, family_data, **kwargs):
        shape = list(family_data.green_strain.shape)
        shape[-1] = shape[-2]
        out = nm.empty(shape, dtype=nm.float64)

        get = family_data.get
        fargs = [get(name, msg_if_none=_msg_missing_data)
                 for name in self.family_data_names]

        self.tan_mod_function(out, mat, *fargs, **kwargs)

        return out

    def get_fargs(self, mat, virtual, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vg, _ = self.get_mapping(state)

        name = state.name
        fd = self.get_family_data(state, self.region, self.integral,
                                  self.geometry_types[name],
                                  self.arg_steps[name],
                                  self.arg_derivatives[name])

        if mode == 'weak':
            if diff_var is None:
                stress = self.compute_stress(mat, fd, **kwargs)
                self.stress_cache = stress
                tan_mod = nm.array([0], ndmin=4, dtype=nm.float64)

                fmode = 0

            else:
                stress = self.stress_cache
                if stress is None:
                    stress = self.compute_stress(mat, fd, **kwargs)

                tan_mod = self.compute_tan_mod(mat, fd, **kwargs)
                fmode = 1

            return (self.weak_function,
                    stress, tan_mod, fd.mtx_f, fd.det_f, vg, fmode,
                    self.hyperelastic_mode)

        elif mode in ('el_avg', 'qp'):
            if term_mode == 'strain':
                out_qp = fd.green_strain

            elif term_mode == 'stress':
                out_qp = self.compute_stress(mat, fd, **kwargs)

            else:
                raise ValueError('unsupported term mode in %s! (%s)'
                                 % (self.name, term_mode))

            fmode = {'el_avg' : 1, 'qp' : 2}[mode]

            return self.integrate, out_qp, vg, fmode

        else:
            raise ValueError('unsupported evaluation mode in %s! (%s)'
                             % (self.name, mode))

    def get_eval_shape(self, mat, virtual, state,
                       mode=None, term_mode=None, diff_var=None, **kwargs):
        from sfepy.mechanics.tensors import dim2sym

        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(state)
        sym = dim2sym(dim)

        if mode != 'qp':
            n_qp = 1

        return (n_el, n_qp, sym, 1), state.dtype

class DeformationGradientTerm(Term):
    r"""
    Deformation gradient :math:`\ull{F}` in quadrature points for
    `term_mode='def_grad'` (default) or the jacobian :math:`J` if
    `term_mode='jacobian'`.

    Supports 'eval', 'el_avg' and 'qp' evaluation modes.

    :Definition:

    .. math::
        \ull{F} = \pdiff{\ul{x}}{\ul{X}}|_{qp}
        = \ull{I} + \pdiff{\ul{u}}{\ul{X}}|_{qp} \;, \\
        \ul{x} = \ul{X} + \ul{u} \;, J = \det{(\ull{F})}

    :Arguments:
        - parameter : :math:`\ul{u}`
    """
    name = 'ev_def_grad'
    arg_types = ('parameter',)
    arg_shapes = {'parameter' : 'D'}

    @staticmethod
    def function(out, vec, vg, econn, term_mode, fmode):
        d = 1 if term_mode == 'jacobian' else vg.dim
        out_qp = nm.empty((out.shape[0], vg.n_qp, d, d), dtype=out.dtype)

        mode = 1 if term_mode == 'jacobian' else 0
        terms.dq_def_grad(out_qp, vec, vg, econn, mode)

        if fmode == 2:
            out[:] = out_qp
            status = 0

        else:
            status = vg.integrate(out, out_qp, fmode)

        return status

    def get_fargs(self, parameter,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        get_saved = False if term_mode == 'act_conf' else True
        vg, _ = self.get_mapping(parameter, get_saved)

        vec = self.get_vector(parameter)

        fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)

        return vec, vg, parameter.field.econn, term_mode, fmode

    def get_eval_shape(self, parameter,
                       mode=None, term_mode=None, diff_var=None, **kwargs):
        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(parameter)

        if mode != 'qp':
            n_qp = 1

        if term_mode == 'jacobian':
            return (n_el, n_qp, 1, 1), parameter.dtype

        else: # 'def_grad'
            return (n_el, n_qp, dim, dim), parameter.dtype



import os
import numpy as nm

from sfepy.base.testing import TestCommon

class Test(TestCommon):
    tests = ['test_log_create', 'test_log_rw']

    @staticmethod
    def from_conf(conf, options):
        return Test(
            conf=conf, options=options,
            log_filename=os.path.join(options.out_dir, 'test_log.txt'),
        )

    def test_log_create(self):
        from sfepy.base.log import Log

        log = Log([['x'], ['x^2', 'x^3']],
                  plot_kwargs = [{},
                                 [{'color' : 'b', 'ls' : '', 'marker' : 'o'},
                                  {'color' : 'g', 'ls' : ':', 'marker' : 'x'}]],
                  yscales=['log', 'linear'],
                  xlabels=['x', 'x'], ylabels=['x', 'x^p'],
                  is_plot=False,
                  aggregate=0, sleep=0.0,
                  log_filename=self.log_filename,
                  formats=[['{:.3e}'], ['{:.5e}'] * 2])

        for x in nm.linspace(0, 1, 11):
            log(x, x**2, x**3, x=[x + 1, x])
            if nm.allclose(x, 0.5):
                log.plot_vlines([0], color='g', linewidth=2)
            if nm.allclose(x, 0.7):
                log.plot_vlines([1], color='g', linewidth=2)

        log(finished=True)

        return True

    def test_log_rw(self):
        from sfepy.base.base import Output
        from sfepy.base.log import read_log, write_log

        log, info = read_log(self.log_filename)

        filename2 = os.path.join(self.options.out_dir, 'test_log2.txt')
        output = Output('', filename=filename2, quiet=True)

        write_log(output, log, info)

        log2, info2 = read_log(filename2)

        ok = True
        _ok = info == info2
        if not _ok:
            self.report('headers are not equal!')
            self.report(info)
            self.report(info2)
        ok = ok and _ok

        for key, val2 in log2.items():
            val = log[key]
            _ok = nm.allclose(val[0], val2[0], rtol=0.0, atol=1e-14)
            if not _ok:
                self.report('x values are not equal!')
                self.report(val[0])
                self.report(val2[0])
            ok = ok and _ok

            _ok = nm.allclose(val[1], val2[1], rtol=0.0, atol=1e-14)
            if not _ok:
                self.report('y values are not equal!')
                self.report(val[1])
                self.report(val2[1])
            ok = ok and _ok

            _ok = nm.allclose(val[2], val2[2], rtol=0.0, atol=1e-14)
            if not _ok:
                self.report('vlines are not equal!')
                self.report(val[2])
                self.report(val2[2])
            ok = ok and _ok

        return ok



from __future__ import absolute_import
from __future__ import print_function
try:
    from .extmods import *
except (ImportError, AttributeError):
    print('sfepy extension modules may not be compiled!\nTry typing "make".')
    raise

from .mesh import Mesh
from .domain import FEDomain
from .fields_base import Field
from sfepy.discrete.fem.meshio import MeshIO
from .utils import extend_cell_data
